"""
Helper functions to generate/manage excels
"""
import io
import pandas as pd

__all__ = [
    'generate_xlsx'
]

def generate_xlsx(spec):
    """
    Generate an excel sheet according to specification

    Args:
        spec (dict): Specification of the excel

    Returns:
        bytes: content of an excel file


    Example of spec::

        [
            {
                "name": "Variables",
                "tables": [
                    {
                        "name": "variables",
                        "position": [0,0],
                        "data": vdf
                    }
                ]
            },        
            {
                "name": "Metrics",
                "tables": [
                    {
                        "name": "metrics",
                        "position": [0,0],
                        "data": mdf
                    }
                ]
            },
        ]
    """

    # Sanity check
    if not isinstance(spec, list) or len(spec) == 0:
        raise Exception("Invalid excel generation specification")

    for sheetspec in spec:
        if not isinstance(sheetspec, dict) or len(sheetspec) == 0:
            raise Exception("Invalid sheet specification. Invalid dictionary")

        if (("name" not in sheetspec) or ("tables" not in sheetspec)):
            raise Exception("Sheet specification should include name of sheet and the tables")

    
    output = io.BytesIO()
    writer = pd.ExcelWriter(output, engine='xlsxwriter')

    for sheetspec in spec:
        sheetname = sheetspec['name']
        for t in sheetspec['tables']:
            name = t['name']
            df= t['data']
            position = t['position']
            extra = t.get('extra', {})
            if len(df) > 0: 
                df.to_excel(writer,
                            sheet_name=sheetname,
                            index=False,
                            encoding='utf-8',
                            startrow=position[0],
                            startcol=position[1],
                            **extra)
                        

    writer.save()
    output.seek(0)
    
    return output.read()

if __name__ == "__main__":

    vdf = pd.DataFrame([
        { 'a': 2, 'b': 2},
        { 'a': 3, 'b': 4}
    ])

    mdf = pd.DataFrame([
        { 'c': 2, 'd': 2},
        { 'c': 3, 'd': 4}
    ])    
    
    excelspec = [
        {
            "name": "Variables",
            "tables": [
                {
                    "name": "variables",
                    "position": [0,0],
                    "data": vdf
                }
            ]
        },        
        {
            "name": "Metrics",
            "tables": [
                {
                    "name": "metrics",
                    "position": [0,0],
                    "data": mdf
                }
            ]
        },
    ]
    content = generate_xlsx(excelspec)
    with open('excelspectest.xlsx', 'wb') as fd:
        fd.write(content)



import io
import re
import os
import time
import json
import hashlib
import sqlite3
import datetime
import pandas as pd
from contextlib import redirect_stdout

from langchain.agents import create_sql_agent
from langchain.agents import create_pandas_dataframe_agent
from langchain.agents.agent_toolkits import SQLDatabaseToolkit
from langchain.sql_database import SQLDatabase
from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI
from langchain.agents import AgentExecutor

from . import agent_events
from llmsdk.lib import SafeEncoder
from llmsdk.services.log import *


class LLMDataQuerier(object):
    """
    Class to do querying of a dataframe using LLMs
    """

    def __init__(self,
                 name,
                 cred,
                 mode,
                 data,
                 cascade,
                 debug=False):
        """
        init the dataframe LLM query agent
        name: name of the agent
        cred: credentials object
        mode: type of input dataset
                'csv': data will be available as path to a csv
                'sqlite': data will be available as path to a sqlite db
                'df': data will be available as pandas dataframe
        data: pointer to data, see mode param
        cascade: list of LLM backends to use
                    the list will be tried in order for each query
                    on failure, the next LLM in the list will be tried
        debug: if True, returns a bunch of useful information for debugging
        """
        start_time = time.time()

        # logging
        self.logger = get_logger()

        # defaults
        self.max_llm_tokens = 1024 # max tokens in the response
        self.mode = mode
        self.data = data
        self.debug = debug

        # name
        self.agent_name = name

        # creds
        self.cred = cred
        # LLM params
        self.cascade = cascade

        # init the llm objects
        self.llms = self._get_llm_objs(cascade=self.cascade, cred=self.cred)

        # init the agents
        self.agents = self._get_agent(mode=self.mode, data=self.data)

        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "cascade": self.cascade,
                "mode": self.mode,
                "data": f"dataframe of shape {self.data.shape}" if self.mode=='df' else self.data,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## helper functions

    def _create_id(self, string):
        """
        create a unique identifier from a string
        """
        return hashlib.md5(string.encode('utf-8')).hexdigest()

    def _log_event(self, event_name, duration=0, params={}):
        """
        format an event object to store in the agent's metadata
        event_name: name of the event to log
        duration: how long did this event take, only needed for events
                   that require some waiting (e.g. RTT to LLM APIs)
        params: any other info needed to be logged
        """

        ts = time.time() - duration # to get the actual start time
        event_id = self._create_id(f'{event_name}-{ts}')

        event = {
            "id": event_id,
            "agent": self.agent_name,
            "timestamp": round(ts, 3),
            "duration": round(duration, 3),
            "name": event_name,
            "params": params,
        }
        self.metadata['events'].append(event)

        return event

    def _get_llm_objs(self, cascade, cred):
        # init the LLM models
        ##
        ## cascade = [{"platform": platform, "model": model}, ...]
        ##

        # run through each entry in the LLM cascade

        llms = []
        for entry in cascade:
            # get the platform name
            _id = entry.get("id")
            if _id is None:
                self.logger.warn("Cannot init LLM in cascade, no ID",
                             extra={
                                 'source': self.agent_name,
                                 'data': entry
                             })
                continue

            # we can proceed
            platform = entry.get("platform")

            if platform == "openai":
                # get the api key from creds
                api_key = self._get_api_key(cred, platform)

                ###
                # langchain indiosyncracy
                # needed for the SQL agent later on
                #
                if platform == 'openai' and api_key is not None:
                    os.environ['OPENAI_API_KEY'] = api_key

                ###

                # get the llm objects
                model = entry.get("model")
                if model in ["gpt-3"]:
                    # non-chat models
                    llm = OpenAI(temperature=0,
                                           model=model,
                                           max_tokens=self.max_llm_tokens,
                                           openai_api_key=api_key)
                elif model in ["gpt-3.5-turbo", "gpt-4"]:
                    # chat models
                    llm = ChatOpenAI(temperature=0,
                                               model=model,
                                               max_tokens=self.max_llm_tokens,
                                               openai_api_key=api_key)
                else:
                    self.logger.warn("Cannot init LLM in cascade, unknown model",
                             extra={
                                 'source': self.agent_name,
                                 'data': entry
                             })
                    continue

                # add the LLM backend to the list
                llm_backend = {
                    "id": _id,
                    "platform": platform,
                    "model": model,
                    "llm": llm
                }
                llms.append(llm_backend)

            else:
                self.logger.warn("Cannot init LLM in cascade, unknown platform",
                             extra={
                                 'source': self.agent_name,
                                 'data': entry
                             })

        if len(llms) == 0:
            raise Exception("No LLMs initialized, cannot proceed")

        return llms

    def _get_api_key(self, cred, key):
        """
        get the API key from the cred
        """
        api_key = None
        if isinstance(cred, str):
            api_key = cred
        if isinstance(cred, dict) and key in cred:
            api_key = cred[key]
        return api_key

    def _construct_sqlagent_prompt(self, data):
        """
        construct prompt for the SQL agent
        give as much context as possible to help the LLM
        """
        # construct table schema
        tables = data.get("tables", [])
        num_tables = len(tables)

        table_stmt = f"The database has {num_tables} table(s). The tables and corresponding columns are as follows:" + "\n"

        for table in tables:
            table_name = table['name']
            table_desc = table['desc']
            cols = table["cols"]

            cols_text = []
            for col in cols:
                col_name = list(col)[0]
                col_desc = col[col_name]
                cols_text.append(f"\t{col_name}: {col_desc}")
            cols_text = ",\n".join(cols_text)

            # add the table context helpers
            table_context = table['context']
            table_context = [f"- {context}" for context in table_context]
            table_context = "\n".join(table_context)

            table_stmt += f"""
TABLE NAME: {table_name}
TABLE DESCRIPTION: {table_desc}
COLUMN names and descriptions:
{cols_text}

Keep the following context in mind when querying the {table_name} table:
{table_context}
"""

        # get the date for the agent
        x = datetime.datetime.now()
        today = x.strftime("%A, %B %d, %Y")

        SQL_PREFIX = f"""You are an agent designed to interact with a SQL database.

{table_stmt}

Your task is as follows:
Given an input question, create a syntactically correct {{dialect}} query to run, then look at the results of the query and return the answer.
You can order the results by a relevant column to return the most interesting examples in the database.
Never query for all the columns from a specific table, only ask for the relevant columns given the question.

You have access to tools for interacting with the database.
Only use the below tools. Only use the information returned by the below tools to construct your final answer.
You MUST double check your query before executing it. If you get an error while executing a query, rewrite the query and try again.

Ensure that you follow these rules:
- DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.
- If the answer contains multiple items, format the answer as a json list.
- If you need to do string comparisons, always use the LOWER() function.
- If you need the date to answer any questions, remember that today is {today}.
- If the question does not seem to be related to the database, just return "I don't know" as the answer.
"""

        return SQL_PREFIX


    def _get_agent(self, mode, data):
        """
        init the agent
        """

        mode = data.get("mode")
        dataset = data.get("dataset")

        # check for valid mode
        if mode not in ['sql', 'csv', 'df']:
            raise Exception("Unsupported mode for input data")

        # setup agent for sql
        if mode == 'sql':
            # setup the DB conn
            db = SQLDatabase.from_uri(f"sqlite:///{dataset}")

            # override LangChain SQL agent prompt prefix
            # so that the agent doesn't use LIMIT statements
            SQL_PREFIX = self._construct_sqlagent_prompt(data)

        # setup agent for dataframes
        elif mode == 'csv':
            # Handle bad quality csv as well..
            df = pd.read_csv(data, on_bad_lines='skip', encoding='unicode_escape')
        elif mode == 'df':
            df = data
        else:
            raise Exception(f"Unsupported mode: {mode}")

        # finally, create one agent per LLM backend
        agents = {}
        for llm_backend in self.llms:
            # get the llm object
            llm = llm_backend['llm']
            # for csv agents
            if mode in ['csv', 'df']:
                agent = create_pandas_dataframe_agent(llm=llm,
                                                        df=df,
                                                        verbose=True)
            # for sql agents
            elif mode in ['sql']:
                toolkit = SQLDatabaseToolkit(db=db, llm=llm)
                agent = create_sql_agent(llm=llm,
                                            toolkit=toolkit,
                                            prefix=SQL_PREFIX,
                                            verbose=True)
            # force intermediate steps on return
            # we use this to get the agent's chain of thought
            agent.return_intermediate_steps = True
            # agent object
            agent_obj = {
                "platform": llm_backend['platform'],
                "model": llm_backend['model'],
                "agent": agent,
            }
            _id = llm_backend['id']
            agents[_id] = agent_obj

        return agents

    def _get_dialect(self, mode=None):
        """
        return the code dialect given the agent's mode
        """
        dialects = {
            "sql": "sql",
            "df": "pandas",
            "csv": "pandas",
        }
        if not mode:
            mode = self.mode
        dialect = dialects.get(mode)

        return dialect

    def _parse_code_from_thought(self, thoughts):
        """
        parse the chain-of-thought information sent by the LLM for the code snippets
        """
        # get the dialect of the code
        dialect = self._get_dialect()

        snippets = []
        # step through each thought, but in reverse order
        # keep storing successful thoughts
        # when we hit the first failure, we break out
        thoughts.reverse()
        for one_step in thoughts:
            # check by dialect
            if dialect == 'sql':
                # get the components
                tool = one_step['tool'].lower()
                tool_input = one_step['tool_input'] # don't lowercase
                observation = one_step['observation'].lower()
                # get the code snippet
                if tool == 'sql_db_query' and "error: (sqlite3.operationalerror)" not in observation:
                    snippets.append(tool_input)
                if tool == 'sql_db_query' and "error: (sqlite3.operationalerror)" in observation:
                    # we have found a failure
                    # most likely, the LLM would have failed
                    # upto this point
                    # so we can break
                    break
            if dialect == 'pandas':
                # get the components
                tool = one_step['tool'].lower()
                tool_input = one_step['tool_input'] # don't lowercase
                # get the code snippet
                if tool == 'python_repl_ast':
                    snippets.append(tool_input)

        thoughts.reverse() # reverse the thoughts list for consistency with input
        snippets.reverse() # we need the reverse order now
        code = {
            "dialect": dialect,
            "snippets":  snippets
        }

        return code

    def _get_chain_of_thought(self, result):
        """
        parse the intermediate steps list
        to construct human-readable chain-of-thought
        of the agent working through its solution
        """
        thoughts = []

        # get the dialect of the code
        dialect = self._get_dialect()

        # get the intermediate steps
        steps = result.get("intermediate_steps", [])

        # foreach step
        for step in steps:
            agent_action = step[0]
            observation = step[1]

            tool = agent_action.tool
            tool_input = agent_action.tool_input
            log = agent_action.log

            # get the thought
            m = None
            if dialect == "sql":
                m = re.search('^(.+?)\s+?Action:', log, flags=re.IGNORECASE)
            elif dialect == "pandas":
                m = re.search('^Thought(.+?)\s+?Action:', log, flags=re.IGNORECASE)
            if m:
                thought = m.group(1).strip()
            else:
                thought = "BEGIN"

            # contruct the ReACT step
            one_step = {
                "thought": thought,
                "tool": tool,
                "tool_input": tool_input,
                "observation": observation,
            }
            thoughts.append(one_step)

        return thoughts

    def _is_json(self, string):
        """
        check if string is json
        """
        try:
            json.loads(string)
        except ValueError as e:
            return False
        return True

    def _err_msg(self, t):
        msgs = {
            "err": "I'm having trouble understanding. Try another way of wording your query."
        }
        return msgs.get(t, "Something went wrong, try your query again.")

    # interfaces

    def get_metadata(self):
        """
        return metadata collected by the agent
        """
        return self.metadata

    def mux_query_one_try(self, query, agent_id, agent):
        """
        do one try of the query through a specified agent in the cascade chain
        """
        # get the llm backend
        agent_exec = agent['agent']

        try:
            # redirect stdout so we can capture the agent's chain-of-thought
            # this is a hack until we figure out how to get chain-of-thought
            # directly from LangChain
            f = io.StringIO()
            with redirect_stdout(f):
                result = agent_exec(query)
            thought = f.getvalue()
            # check if LLM could not find solution
            if result['output'].lower() == "i don't know.":
                raise Exception("Force exception")
        except:
            # something went wrong when trying the query
            result = { "input": query, "output": self._err_msg('err')}
            thought = ""
            success = False
        else:
            success = True
            result['cascade'] = { "id": agent_id }
            for key in ['platform', 'model']:
                result['cascade'][key] = agent[key]


        return success, result, thought


    def query(self, query, cascade_id=None):
        """
        run a query on the dataframe
        query: query string
        cascade_id: id of the model to use in the cascade list specified during agent init
        """
        start_time = time.time()

        seq = 0
        tries = []
        if cascade_id == None:
            # we have not specified an agent id
            # in the cascade list so try them all
            for _id, agent in self.agents.items():
                # try one agent in the cascade
                success, result, thought = self.mux_query_one_try(query, _id, agent)
                tries.append({"seq": seq, "cascade_id": _id, "success": success})
                seq += 1
                if success:
                    break
        else:
            # we have specified an agent id
            # in the cascade list so try only that
            agent = self.agents.get(cascade_id)
            if not agent:
                raise Exception(f"Unknown agent ID {cascade_id}, not specified in cascade")

            # try the requested agent in the cascade
            success, result, thought = self.mux_query_one_try(query, cascade_id, agent)
            tries.append({"seq": seq, "cascade_id": cascade_id, "success": success})

        result['success'] = success
        result['tries'] = tries

        # format the result keys
        if result.get('input'):
            result['query'] = result.pop('input')
        if result.get('output'):
            result['answer'] = result.pop('output')
        result['intermediate_steps'] = [] if 'intermediate_steps' not in result else result['intermediate_steps']

        # check for result type
        is_json = self._is_json(result['answer'])
        if is_json:
            result['answer'] = json.loads(result['answer'])
            result['type'] = 'json'
        else:
            result['type'] = 'str'

        # get the chain of thought
        # strip ANSI control sequences first
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        thought = ansi_escape.sub('', thought)
        # get line-by-line chain of thought from stdout
        result['raw_thoughts'] = thought.split("\n")

        # convert intermediate steps to chain of thought
        # this is more stable
        result['chain_of_thought'] = self._get_chain_of_thought(result)

        # get the code snippets from the chain of thought
        # result['code'] = self._parse_code_from_rawthought(result['raw_thoughts'])
        result['code'] = self._parse_code_from_thought(result['chain_of_thought'])

        # log the event
        params = {
            "query": query,
            "cascade_id": cascade_id,
            "result": result.copy() if isinstance(result, dict) else result
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # decide on what fields to return
        # we do it after the logging so that logs have everything
        if not self.debug:
            r = result.pop('raw_thoughts')
            r = result.pop('intermediate_steps')

        return result




if __name__ == "__main__":

    cred = None
    customer = "acme_2"
    # dbpath = '.../file.sqlite'    # if using SQLite DB
    # csvpath = '.../file.csv'      # if using CSV
    # df = pd.read_csv(csvpath)     # if using raw Pandas DF
    # csvpath = os.path.expandvars("$DATA_ROOT/$AGENTNAME/acme_1/revenue.dashboard_market.csv")
    # df = pd.read_csv(csvpath)     # if using raw Pandas DF

    def get_llm_cascade():
        llm_cascade = [
            {"id": "economy", "platform": "openai", "model": "gpt-3.5-turbo"},
            {"id": "power", "platform": "openai", "model": "gpt-4"},
        ]
        return llm_cascade

    def get_dataset(customer, dbpath):
        data = {
            "acme_1": {
                "mode": "sql",
                "dataset": dbpath,
                "tables": [
                    {
                        "name": "disha",
                        "desc": "contains information about currency transfers between partners across countries",
                        "context": [
                            "each row in the table includes multiple transactions",
                            "when querying for month columns, always use a two-digit month number (e.g. June will be 06 instead of 6)"
                        ],
                        "cols": [
                            {"year": "year during which the transaction occured"},
                            {"month": "two-digit month number in which the transaction occured"},
                            {"source_partner": "name of the partner who sent the transaction"},
                            {"destination_partner": "name of the partner who received the transaction"},
                            {"source_country": "country from which the transaction was sent"},
                            {"destination_country": "country in which the transaction was received"},
                            {"financial_institution": "name of the financial institution that sent the transaction"},
                            {"destination_currency": "currency code in which the transaction occured"},
                            {"net_sales": "total net sales volume"},
                            {"net_revenue": "total net revenue"},
                            {"successful_txns": "total number of successful transactions"},
                            {"failed_txns": "total number of failed transactions"},
                            {"is_new_partner": "binary field indicating whether the source partner is a new partner"},
                            {"is_new_country": "binary field indicating whether the source country is a new country"},
                            {"b2b": "binary field indicating whether the source partner is a B2B partner"}
                        ]
                    }
                ]
            },
            "acme_2": {
                "mode": "sql",
                "dataset": dbpath,
                "tables": [
                    {
                        "name": "surveyresults",
                        "desc": "contains information on sentiment scores for products across various segments of survey respondents",
                        "context": [
                            "a segment is defined as the combination of the country, gender, and age_range columns",
                            "when asked to provide interesting insights, respond with the following statistics in bullet point format - top 3 and bottom 3 segments, product with highest sentiment score, product with highest sentiment score among women but lowest sentiment score among men",
                            "when asked to summarize the dataset, provide your response in bullet point format"
                        ],
                        "cols": [
                            {"product_id": "system identifier for product being surveyed"},
                            {"country": "country to which survey respondent belongs"},
                            {"gender": "gender of respondent (female, male)"},
                            {"age_range": "age range bucket into which survey respondent falls (under 18, 18 to 24, 25 to 34, 35 to 44, 45 to 54, above 55)"},
                            {"sentiment_score": "integer number indicating sentiment or affinity of a segment of survey respondents to the product (range is 50 to 100, higher score are more positive sentiment)"},
                            {"le_score": "float number indicating line efficiency score of a segment of survey respondents to the product"},
                            {"overall_score": "float number indicating importance score of a product"},
                        ]
                    }
                ]
            },
        }
        return data[customer]

    # (query, cascade_id)
    queries = {
        "acme_1": [
            ("how many transactions were made this year?", None),
            ("how many transactions were made this year?", "power"),
            ("from how many countries were txns sent this month last year?", None),
            ("how many partners have been serviced this month and year?", None),
            ("how many partners have been serviced this month and year?", "power"),
            ("who are all the new source partners this year", "economy"),
            ("who are all the new source partners this year", None),
            ("what was the total revenue generated from new partners this year", None),
            ("list the top 10 financial institutions by revenue sending transactions to Egypt", None),
            ("list all new b2b partners this year", None),
            ("list all partners in India", None)
        ],
        "acme_2": [
            ("what product and segment shows the highest sentiment?", None),
            ("what are the top 3 products in the US for women between 18 and 24?", None),
            ("what is the top segment for each product?", None),
            ("what is the top product in the US for older women?", None),
            ("what are some interesting insights on this dataset?", "power")
        ],
    }

    # get the dataset spec
    dataset = get_dataset(customer, dbpath)
    # get the cascade chain spec
    cascade = get_llm_cascade()

    # setup the agent and ask questions
    agent = LLMDataQuerier(name="datagpt",
                           cred=cred,
                           mode='sql',
                           data=dataset,
                           cascade=cascade,
                           debug=False)

    # answers
    for entry in queries[customer]:
        query = entry[0]
        cascade_id = entry[1]
        result = agent.query(query=query, cascade_id=cascade_id)
        print(f"QUESTION (cascade_id={cascade_id}): {query}")
        print(f"ANSWER: {result.get('answer')}")
        print(f"METADATA: {json.dumps(result, indent=4, cls=SafeEncoder)}")
        print("\n----------\n")



"""
logging
~~~~~~~

This module contains a class that wraps the log4j object instantiated
by the active SparkContext, enabling Log4j logging for PySpark using.
"""


class Log4j(object):
    """Wrapper class for Log4j JVM object.

    :param spark: SparkSession object.
    """

    def __init__(self, spark):
        # get spark app details with which to prefix all messages
        app_id = spark.sparkContext.getConf().get("spark.app.id")
        app_name = spark.sparkContext.getConf().get("spark.app.name")

        log4j = spark._jvm.org.apache.log4j
        message_prefix = "<" + app_name + " " + app_id + ">"
        self.logger = log4j.LogManager.getLogger(message_prefix)

    def error(self, message):
        """Log an error.

        :param: Error message to write to log
        :return: None
        """
        self.logger.error(message)
        return None

    def warn(self, message):
        """Log an warning.

        :param: Error message to write to log
        :return: None
        """
        self.logger.warn(message)
        return None

    def info(self, message):
        """Log information.

        :param: Information message to write to log
        :return: None
        """
        self.logger.info(message)
        return None



# https://stackoverflow.com/questions/62894952/fastapi-gunicon-uvicorn-access-log-format-customization
import os
import logging

if 'LOGDIR' in os.environ:
    APP_LOG_FILE = os.path.join(os.environ['LOGDIR'],
                                os.environ['AGENTNAME'],
                                "app.json.log")
    try:
        os.makedirs(os.path.dirname(APP_LOG_FILE))
    except:
        pass
    applogconfig = {
        "level": "DEBUG",
        "class": "logging.handlers.RotatingFileHandler",
        "filename": APP_LOG_FILE,
        "maxBytes": 1024 * 1024 * 10,  # 10
        "backupCount": 100,
        "formatter": "json",
    }
else:
    applogconfig = {
        "formatter": "default",
        "class": "logging.StreamHandler",
        "stream": "ext://sys.stdout",
    }


log_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "json": {
            "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
            "format": "%(message)%(levelname)%(name)%(asctime)%(funcName)%(lineno)%(pathname)%(module)%(created)",
        },
        'access': {
            '()': 'uvicorn.logging.AccessFormatter',
            'fmt': '%(levelprefix)s %(asctime)s - %(client_addr)s - "%(request_line)s" %(status_code)s',
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "use_colors": False
        },
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(levelprefix)s %(asctime)s - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "use_colors": False
        },
    },
    "handlers": {
        "app": applogconfig,
        'access': {
            'class': 'logging.StreamHandler',
            'formatter': 'access',
            'stream': 'ext://sys.stdout'
        },
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
    },
    "loggers": {

        "app": {
            "handlers": ["app"],
            "level": "DEBUG",
            "propagate": False
        },
        "uvicorn": {
            "handlers": ["access"],
            "level": "DEBUG",
            "propagate": True
        },
        'uvicorn.access': {
            'handlers': ['access'],
            'level': 'INFO',
            'propagate': False
        },
        'uvicorn.error': {
            'level': 'INFO',
            'propagate': False
        }
    },
}


def get_logger():
    return logging.getLogger("app")



import os
import sys
import json
from enrichsdk import Source
import logging

logger = logging.getLogger("app")


class JSONSource(Source):
    """
    Load a file into a 'dict' frame in the state.

    Params are meant to be passed as parameter to update_frame.

    Example configuration::

         ...

         "args": {
             "source": {
                 'hello': {
                     'frametype': 'dict',
                     'filename': '%(data_root)s/shared/hello.json',
                     'params': {}
                 }
             }
         }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "JSONSource"

        self.testdata = {
            "conf": {
                "args": {
                    "source": {
                        "hello": {
                            "filename": "%(data_root)s/shared/hello.json",
                            "frametype": "dict",
                            "params": {},
                        }
                    }
                }
            },
            "data": {},
        }

    def preload_clean_args(self, args):
        """
        Check if the args are consistent with the
        specification.
        """

        # Backward compatabilty
        if "source" not in args:
            args = {"source": args}

        args = super().preload_clean_args(args)

        # Sanity check...
        assert isinstance(args, dict)
        assert "source" in args
        assert isinstance(args["source"], dict)

        source = args["source"]
        for name, detail in source.items():

            if ("frametype" not in detail) or (detail["frametype"] != "dict"):
                logger.error(
                    "Invalid configuration. Only JSON source supported by this source transform",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            if (
                ("filename" not in detail)
                or (not isinstance(detail["filename"], str))
                or ("params" not in detail)
                or (not isinstance(detail["params"], dict))
            ):
                logger.error(
                    "Invalid args. Filename (string) and params (dict) are required",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            filename = detail["filename"]
            if not filename.lower().endswith(".json"):
                logger.error(
                    "Input file must a .json file",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            # => Materialize the path...
            detail["filename"] = self.config.get_file(detail["filename"])

        return args

    def validate_args(self, what, state):
        """
        Double check the arguments
        """
        assert isinstance(self.args, dict)
        assert "source" in self.args
        assert isinstance(self.args["source"], dict)

        source = self.args["source"]
        for name, detail in source.items():
            assert ("frametype" in detail) and (detail["frametype"] == "dict")
            assert "filename" in detail
            assert "params" in detail

    def process(self, state):
        """
        Load the json files into 'dict' frames and store them in the state.
        """
        logger.debug(
            "{} - process".format(self.name),
            extra=self.config.get_extra({"transform": self.name}),
        )

        source = self.args["source"]
        for name, detail in source.items():

            filename = detail["filename"]
            data = json.load(open(filename))

            updated_detail = {
                "df": data,
                "transform": self.name,
                "frametype": "dict",
                "params": [
                    {
                        "type": "compute",
                    },
                    {
                        "type": "lineage",
                        "transform": self.name,
                        "dependencies": [
                            {"type": "file", "nature": "input", "objects": [filename]}
                        ],
                    },
                ],
                "history": [
                    # Add a log entry describing the change
                    {"transform": self.name, "log": "Loaded json file"}
                ],
            }

            # Update the state.
            state.update_frame(name, updated_detail, create=True)

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """

        source = self.args["source"]
        for name, detail in source.items():
            if not state.reached_stage(name, self.name):
                raise Exception("Could not find new frame created for {}".format(name))

                detail = state.get_frame(name)
                df = detail["df"]

                # Check if it is a valid dictionary...
                assert isinstance(df, dict)


provider = JSONSource



import os
import sys
import gzip
import csv
import io
import re
import time
from celery import shared_task
from datetime import datetime, timedelta
from dateutil import relativedelta
from celery.utils.log import get_task_logger

from django.db import connections

logger = get_task_logger(__name__)


@shared_task(name="complex_task")
def complex_task(params):
    """
    Run some complicated and long running transaction
    """

    try:
        records = [{"name": "John", "location": "South Africa"}]

        logger.info(f"Received {len(records) -1} records")
    except:
        logger.exception("Error while executing the query")
        raise

    return {"records": records}



import time
import json
import string

from langchain.schema import AIMessage, HumanMessage, SystemMessage
from langchain.callbacks import get_openai_callback

from llmsdk.agents.basellm import BaseLLMAgent

from . import agent_events
from llmsdk.lib import SafeEncoder
from llmsdk.services.log import *

__all__ = ['LLMQueryMapper']


class LLMQueryMapper(BaseLLMAgent):
    """
    Class to do querying of a set of strings to return the closest matching string to the query
    """

    def __init__(self,
                 name,
                 cred={},
                 platform="openai",
                 statestore="redis",
                 topk=5):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        platform: name of the LLM platform backend to use
                default to OpenAI GPT platform for now, Azure is also supported
                will be extended in the future to suuport other models
        memory_size: how many tokens of memory to use when chatting with the LLM
        """

        start_time = time.time()

        # init the base class
        super().__init__(name=name,
                         cred=cred,
                         platform=platform,
                         agent_type="querymapper",
                         statestore=statestore)

        # defaults
        self.chunk_size = 1000
        self.chunk_overlap = 300
        self.index = None
        self.metadata = {}
        self.vdb_client = None
        self.index_name = None
        self.index_store = None
        self.topk = topk
        self.doc_signatures = []
        self.docs = {}

        # LLM params
        self.platform = platform
        self.chaintype = "stuff"

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                       cred=self.cred)

        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "type": self.agent_type,
                "platform": self.platform,
                "chaintype": self.chaintype,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## helper functions

    def _get_query_prompt_ranking(self, query, docs, spec):
        """
        generate a prompt for running a query to find the most relevant doc from a set
        """
        # how many docs to choose between
        num_choices = len(docs)

        # construct the choices
        choices = []
        for d in docs:
            choice = f"TEXT: {d.page_content}" + "\n"
            choice += f"ID: {d.metadata['id']}" + "\n"
            choice += f"SCORE: {1-d.metadata['distance']}"
            choices.append(choice)
        choices = "\n\n".join(choices)

        # define output format
        output_format = "Respond with a json dictionary as follows: " + "\n"
        output_format += '{"is_similar": is_similar, id": ID}'

        # construct the context
        context = spec.get("context", [])
        context = [f'- {c}' for c in context]
        context = "\n".join(context)

        # construct the prompt
        sys_msg = f"""You are a highly advanced AI-enabled semantic text matching program.

You will be presented with a user's query and a set of {num_choices} possible matching text chunks.
The text chunks will be presented as successive sets of:
TEXT: text_chunk
ID: identifier
SCORE: relevance_score

You task is to respond with two pieces of information: 'is_similar' and 'id'

First, determine whether the query is similar to any of the text chunks.
If the query is similar, then set 'is_similar' to true else set it to false

Second, determine the ID of the MOST RELEVANT text chunk that matches the query.
Use your judgement, the text chunk with the highest SCORE may or may not always be the most relevant match.

{output_format}

Keep the following context in mind while performing your task:
{context}
"""
        human_msg = f"""
------ BEGIN QUERY ------
{query}
------ END QUERY ------

------ BEGIN TEXT CHUNKS ------
{choices}
------ END TEXT CHUNKS ------

Response:"""

        messages = [
            SystemMessage(content=sys_msg),
            HumanMessage(content=human_msg),
        ]

        return messages

    def _get_query_prompt_intent(self, query, answer, spec):
        """
        generate a prompt for extracting intent params from a query template
        """

        metadata = answer.get("selection", {}).get("metadata", {})

        # get intent variables
        key = metadata.get("key", "")
        intents = key.split('|')[1:]
        intents = list(set(intents))
        intents_str = [f'- {i}' for i in intents]
        intents_str = "\n".join(intents_str)

        # get the context
        context = spec.get("context", [])
        context = [f'- {c}' for c in context]
        context = "\n".join(context)

        # get the rules
        rules = json.loads(metadata.get("rules", json.dumps([])))
        rules = [f'- {r}' for r in rules]
        rules = "\n".join(rules)

        # define output format
        op = [f'"{i}": value' for i in intents]
        output_format = "{" + ", ".join(op) + "}"

        # construct the prompt
        sys_msg = f"""You are a highly advanced AI-enabled intent detection program.

You will be presented with a user's query and a set of intent variables to identify from the query.
The intent variables are:
{intents_str}

Your task is to extract the value of each intent variable from the query and respond.

Respond with ONLY the intent key and values in a json dictionary as follows:
{output_format}

Keep the following context in mind while performing your task:
{context}
{rules}
"""
        human_msg = f"""
------ BEGIN QUERY ------
{query}
------ END QUERY ------

Response:"""

        messages = [
            SystemMessage(content=sys_msg),
            HumanMessage(content=human_msg),
        ]

        return messages


    ##
    ## interfaces
    ##

    def run_query_ranking(self, query, spec):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers using a private source of data
        """
        # get the similar docs
        docs = self.get_similar_docs(query, topk=self.topk)
        sources = [{"id": d.metadata.get('id'), "content": d.page_content, "metadata": d.metadata, "distance": d.metadata.get('distance')} for d in docs]

        # construct the prompt
        prompt = self._get_query_prompt_ranking(query, docs, spec)

        # call the LLM
        response = self.llm(prompt)

        # construct result
        is_relevant = False
        suggestions = []
        try:
            answer = json.loads(response.content)
            _id = answer.get("id")
            # we refer to this as relvance from here on
            is_relevant = answer.get("is_similar")
            if is_relevant:
                for s in sources:
                    if _id == s.get("metadata", {}).get("id"):
                        answer["selection"] = s
                        break
        except:
            pass

        if is_relevant and "selection" not in answer:
            # we are defaulting to the most relevant match according to semantic search
            answer["selection"] = sources[0]
            answer["id"] = sources[0].get("id")
            suggestions = [d.page_content for d in docs]

        if not is_relevant:
            suggestions = [d.page_content for d in docs]

        # run the query against the similar docs
        result = {
            "success": True,
            "is_relevant": is_relevant,
            "question": query,
            "answer": answer,
            "suggestions": suggestions,
            "sources": sources,
        }

        return result

    def run_query_intent(self, query, answer, spec):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers using a private source of data
        """
        # construct the prompt
        prompt = self._get_query_prompt_intent(query, answer, spec)

        # call the LLM
        response = self.llm(prompt)

        # construct result
        try:
            answer = json.loads(response.content)
            success = True
        except:
            success = False

        # run the query against the similar docs
        result = {
            "success": success,
            "question": query,
            "answer": answer,
        }

        return result

    def query(self, query, spec):
        """
        run a query on an index using an llm chain object
        query: query string
        """

        start_time = time.time()

        # first, get the top-ranked query template
        # run the ranking query
        try:
            if self.platform in ['openai', 'azure']:
                with get_openai_callback() as cb:
                    result_ranking = self.run_query_ranking(query, spec)
                stats = {
                    "total_tokens": cb.total_tokens,
                    "prompt_tokens": cb.prompt_tokens,
                    "completion_tokens": cb.completion_tokens,
                    "total_cost": round(cb.total_cost, 4)
                }
            else:
                result_ranking = self.run_query_ranking(query, spec)
                stats = {}
        except:
            result_ranking = {
                "success": False,
                "question": query,
                "answer": self._err_msg('field'),
                "sources": [],
            }
            stats = {}

        # log the event
        params = {
            "query": query,
            "mode": "ranking",
            "result": result_ranking.copy() if result_ranking is not None else None,
            "stats": stats,
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # for the final return
        result = result_ranking

        # then, parse the query template to get the widget key
        if result_ranking.get("success") and result_ranking.get("is_relevant"):
            # start timing again
            start_time_intent = time.time()

            # run the extraction query
            ranking_answer = result_ranking.get("answer", {})
            try:
                if self.platform in ['openai', 'azure']:
                    with get_openai_callback() as cb:
                        result_intent = self.run_query_intent(query, ranking_answer, spec)
                    stats = {
                        "total_tokens": cb.total_tokens,
                        "prompt_tokens": cb.prompt_tokens,
                        "completion_tokens": cb.completion_tokens,
                        "total_cost": round(cb.total_cost, 4)
                    }
                else:
                    result_intent = self.run_query_intent(query, ranking_answer, spec)
                    stats = {}
            except:
                result_intent = {
                    "success": False,
                    "question": query,
                    "answer": self._err_msg('field'),
                    "sources": [],
                }
                stats = {}

            # construct the key
            if result_intent['success']:
                intent_answer = result_intent.pop("answer", {})
                key = ranking_answer.get("selection", {}).get("metadata", {}).get("key", "")
                for intent, value in intent_answer.items():
                    if value:
                        key = key.replace(intent, value)
                result_intent["key"] = key

            # log the event
            params = {
                "query": query,
                "mode": "intent",
                "result": result_intent.copy() if result is not None else None,
                "stats": stats,
            }
            duration = time.time() - start_time_intent
            event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

            # for the final return
            result = result_intent
        else:
            result.pop("answer")
            result.pop("sources")

        # add the event to the result
        result['metadata'] = {
            "timestamp": event['timestamp'],
            "duration": time.time() - start_time,
        }

        return result

    def load_spec(self, spec, persist_directory, index_name):
        """
        load up the spec with templates for the agent to refer to
        """

        # make note of the topk relevant expected results
        topk = spec.get("topk", 5)
        context = spec.get("context", [])

        # load the data
        templates = spec.get("templates", [])
        for doc in templates:
            # one search doc at a time
            template = doc.get("template")
            if not template:
                continue

            # add the rules to the metadata
            metadata = doc.get("metadata", {})
            if "rules" in metadata:
                metadata["rules"] = json.dumps(metadata["rules"])

            data = self.load_data(content=template,
                                   source="str",
                                   metadata=metadata)
            if not self.index:
                self.create_add_index(data=data,
                                       store="chroma",
                                       persist_directory=persist_directory,
                                       index_name=index_name)
            else:
                self.add_to_index(data=data)

        # set the topk value for the agent
        # this is done here to handle a peculiarity with the LLMBase class
        self.topk = topk


## TESTBED ##

if __name__ == "__main__":

    def get_profilespec():

        # ruleset
        RULE__TIME_PERIOD = "TIME_PERIOD must be one of 'day', 'week', 'month', 'quarter', or 'year'. Previous three months should be interpreted as a quarter."
        RULE__COUNTRY = "All COUNTRY names must be converted to 2 character ISO country codes"
        RULE__METRIC = "METRIC must be one of 'NSV' or 'NR'"

        # setup the template spec
        default = {
            "name": "acme_gpt",
            "topk": 5,
            "context": [
                "these queries are about cross-border financial transactions (traffic) at a fintech company",
                "'financial institutions' are also known as banks",
                "GSV is a metric measuring sales volume",
                "NSV is a metric measuring sales volume",
                "NR is a metric measuring revenue"
            ],
            "templates": [
                {
                    "template": "how many partners went live over the past TIME_PERIOD",
                    "metadata": {
                        "key": "new_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what new B2B partners went live over the past TIME_PERIOD",
                    "metadata": {
                        "key": "new_b2b_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "which new source and/or destination countries have been launched this TIME_PERIOD",
                    "metadata": {
                        "key": "new_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what is the trend of SOURCE_COUNTRY to DESTINATION_COUNTRY traffic over the last TIME_PERIOD",
                    "metadata": {
                        "key": "traffic_trend|SOURCE_COUNTRY|DESTINATION_COUNTRY|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD,
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of all partners in COUNTRY",
                    "metadata": {
                        "key": "partners|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of all source partners in COUNTRY",
                    "metadata": {
                        "key": "source_partners|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of all destination partners in COUNTRY",
                    "metadata": {
                        "key": "dest_partners|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of 'financial institutions' in COUNTRY",
                    "metadata": {
                        "key": "banks|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "how many senders and/or receivers have been serviced this TIME_PERIOD",
                    "metadata": {
                        "key": "active_sndrs_rcvrs|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what is the expected/projected sales metrics for this TIME_PERIOD",
                    "metadata": {
                        "key": "expected_sales|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the total sales for this TIME_PERIOD",
                    "metadata": {
                        "key": "current_sales|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what is the total revenue for this TIME_PERIOD",
                    "metadata": {
                        "key": "current_sales|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "rank destination countries based on METRIC from SOURCE_COUNTRY",
                    "metadata": {
                        "key": "rank_dest_countries_from_source_country|METRIC|SOURCE_COUNTRY",
                        "rules": [
                            RULE__METRIC,
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "rank source partners based on METRIC from SOURCE_COUNTRY",
                    "metadata": {
                        "key": "rank_source_partners_from_source_country|METRIC|SOURCE_COUNTRY",
                        "rules": [
                            RULE__METRIC,
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what was the sales volumes generated from new partners this TIME_PERIOD",
                    "metadata": {
                        "key": "sales_new_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "from how many countries were transactions received this TIME_PERIOD",
                    "metadata": {
                        "key": "source_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "to how many countries were transactions sent this TIME_PERIOD",
                    "metadata": {
                        "key": "dest_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the highlights of business performance this TIME_PERIOD vs. last TIME_PERIOD",
                    "metadata": {
                        "key": "highlights|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the destination countries which were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_destination_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the destination partners which were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_destination_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the source countries that were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_source_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the source partners who were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_source_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
            ]
        }

        return default

    # test query set
    queries = [
        "what is the trend of traffic over the previous year going from america to ghana?",
        "show me traffic growth from england to india past year?",
        "traffic trend, uk to in, last year?",
        "what are the banks in Korea?",
        "what was the NR, NSV generated from new partners this year?",
        "to which countries were txns sent this year?",
        "summarize the business highlights for the year",
        "what is the total GSV this year?",
        "what is the total NSV this year?",
        "what is the total revenue this year?",
        "how many new partners were added last quarter?",
        "how many tigers were at the zoo?",
        "list the destination countries which had traffic previously, but not in the last month",
        "give me projected NR and NSV this month?",
    ]

    # vars
    index_name = "acme_gpt_index"
    persist_directory = "chromadb123"
    agent_name = "acme_gpt"

    # create an agent
    agent = LLMQueryMapper(name=agent_name,
                            platform="openai")

    # load spec
    spec = get_profilespec()
    agent.load_spec(spec=spec,
                    persist_directory=persist_directory,
                    index_name=index_name)

    # test the queries
    for query in queries:
        result = agent.query(query, spec)
        if result.get("success"):
            key = result.get("key")
            print (f"Query: {query}\nKEY: {key}\n")
        else:
            suggestions = result.get("suggestions")
            print (f"Query: {query}\nSuggestions: {suggestions}\n")



"""
Spark Streaming
----------------

Direct and Regular spark streaming are supported

.. warning::
   One major limitation of the implementation is that one topic
   or topic group is supported.

"""
import logging
import json
from pyspark import SparkContext, SparkConf
from pyspark.streaming import StreamingContext
from pyspark.streaming.kafka import KafkaUtils

from .messaging import MessagingBase

logger = logging.getLogger("app")


class SparkKafkaDirectMessaging(MessagingBase):
    """
    Implement sparkstreaming+kafka (DirectStream)

    Example::

        mg = SparkKafkaDirectMessaging({
            "hosts": ["127.0.0.1:9092", "127.0.0.1:9093"],
            "broker": "127.0.0.1:2181",
            "jars": ["spark-streaming-kafka-0-8-assembly_2.11-2.2.0.jar"],
            "batchDuration": 2,
        mg.connect()
        def handler(params, data):
            ....
        mg.consume('orders', handler, {'region': 'south'})

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if ("hosts" not in self.config) or ("broker" not in self.config):
            raise Exception(
                "Kafka hosts and broker must be specified must be specified"
            )

        # if 'jars' not in self.config:
        #    raise Exception("kafka jars have to be specified")

        if ("jars" in self.config) and (not isinstance(self.config["jars"], list)):
            raise Exception("kafka jars have to be a lsit")

        if ("batchDuration" not in self.config) or (
            not isinstance(self.config["batchDuration"], int)
        ):
            raise Exception("Spark streaming batch duration should be specified (int)")

    def connect(self):

        conf = SparkConf()
        if "jars" in self.config:
            jars = self.config["jars"]
            conf.set("spark.jars", ",".join(jars))

        batchDuration = self.config["batchDuration"]

        sc = SparkContext(appName="kafka", conf=conf)
        self.ssc = StreamingContext(sc, batchDuration=batchDuration)

    def consume(self, topic, callback, params):

        ssc = self.ssc

        hosts = self.config["hosts"]

        params = self.config.get("consume_params", {})

        directKafkaStream = KafkaUtils.createDirectStream(
            ssc,
            [topic],
            kafkaParams={"metadata.broker.list": ",".join(hosts)},
            **params
        )

        handler = lambda rdd: callback(params, rdd)
        directKafkaStream.foreachRDD(handler)

        try:
            ssc.start()
            ssc.awaitTermination()
        finally:
            ssc.stop()


class SparkKafkaMessaging(SparkKafkaDirectMessaging):
    """
    Implement sparkstreaming+kafka (with WAL)

    Example::

        mg = SparkKafkaMessaging({
            "hosts": ["127.0.0.1:9092", "127.0.0.1:9093"],
            "broker": "127.0.0.1:2181",
            "jars": ["spark-streaming-kafka-0-8-assembly_2.11-2.2.0.jar"],
            "batchDuration": 2,
            "consume_params": {
                 "groupId": "raw-event-streaming-consumer"
            })
        mg.connect()
        def handler(params, data):
            ....
        mg.consume('orders', handler, {'region': 'south'})

    """

    def consume(self, topic, callback, params):

        ssc = self.ssc

        broker = self.config["broker"]

        if isinstance(topic, str):
            topics = {topic: 1}
        elif isinstance(topic, dict):
            topics = topic
        else:
            raise Exception("topic must be a string or a dictionary")

        params = self.config.get("consume_params", {})

        kafkaStream = KafkaUtils.createStream(
            ssc, zkQuorum=broker, topics=topics, **params
        )
        # Call for each rdd obtained...
        handler = lambda rdd: callback(params, rdd)
        kafkaStream.foreachRDD(handler)

        try:
            ssc.start()
            ssc.awaitTermination()
        except:
            logger.exception("Terminated the stream")
        finally:
            ssc.stop()



import os
import json
from json import JSONEncoder
import pickle
import numpy as np
import pandas as pd
import time
import sqlite3
import tempfile
import hashlib
import builtins

from datetime import datetime, timedelta, date
import logging
from sqlalchemy import create_engine, text as satext

from dateutil import parser as dateparser

from enrichsdk import Compute, S3Mixin
from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec
from enrichsdk.datasets import TransformDoodle
from enrichsdk.utils import SafeEncoder

logger = logging.getLogger("app")

class DataSanitizerBase(Compute):
    """
    Sanitize data based on rules.

    Features of transform baseclass include:
        * Flexible configuration
        * Highlevel specification of transformations
            * specified data source
            * custom defined rules
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "DataSanitizerBase"
        self.description = "Sanitize data based on rules"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

    @classmethod
    def instantiable(cls):
        return False

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}

    ###########################################
    # Process the transformation spec
    ###########################################
    def process_spec(self, spec, df):

        if df is None:
            logger.exception(
                f"Spec: {spec['name']} -- No data",
                extra={"transform": self.name}
            )
            return None

        transformations = spec['transformations']

        # Now apply the transformations
        msg = ""
        for i, t in enumerate(transformations):
            name = f"xform{i}"
            if not isinstance(t, dict):
                msg += f"[{name}] Skipped. Not a dict\n"
                continue
            name = t.get('name', name)
            method = t.get('method')
            if method is None:
                msg += f"[{name}] No method specified \n"
                continue
            params = t.get('params', {})
            if callable(method):
                func = method
            elif hasattr(self, method):
                func = getattr(self, method)
            elif method in builtins:
                func = getattr(builtin, method)
            else:
                msg += f"[{name}] Invalid method\n"
                continue

            nature = t.get('nature', 'column')
            if nature == "column": 
                columns = t.get('columns', [])
                if len(columns) == 0:
                    msg += f"[{name}] No columns specified\n"
                    continue
                for c in columns:
                    if c not in df.columns:
                        msg += f"[{name}] {c} - Invalid column\n"
                        continue
                    try:
                        df[c] = func(df[c], params)
                        msg += f"[{name}] {c} Done\n"
                    except Exception as e:
                        msg += f"[{name}] {c} - Error! str(e)\n"
                        continue
            else:
                msg += f"[{name}] Unsupported nature: {nature}\n"
                continue
            
        logger.debug(
            f"Processed transformation: {spec['name']}",
            extra={"transform": self.name, "data": msg}
        )

        lineage = {
            "type": "lineage",
            "transform": self.name,
            "dependencies": [
                {
                    "type": "dataframe",
                    "nature": "input",
                    "objects": [spec['name']] 
                    },
                ],
            }        
        self.update_frame(spec['name']+"-transformed", f"Transformed Dataset: {spec['name']}", df, lineage)
        
        return df, msg


    ###########################################
    # Helper Functions
    ###########################################

    def update_frame(self, source, description, df, lineage=None):
        if isinstance(source, str):
            name = source
        else:
            name = source["name"]

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    ###########################################
    # I/O Functions
    ###########################################

    def read_s3_data(self, filename, params={}):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            try:
                df = pd.read_csv(s3.open(filename), **params)
                return df
            except:
                pass
        return None

    def get_dataset_s3(self, spec, source, paths, start_date, end_date):
        '''
        Gets all files from paths and puts them together
        into a single dataframe. If self.args['cache']==True,
        then this consolidated dataframe is cached / read from cache
        as applicable.
        '''
        msg = ""

        run_date    = self.args['run_date']
        dataset     = source['dataset']
        params      = source.get('params', {})

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{start_date}-to-{end_date}"
        cachefile = f"cache/{self.name}-rawdata-cache-" + cachename + ".csv"

        # read from cache if available
        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile, **params)
                logger.debug(f"Read cached {dataset}", extra={"transform": self.name, "data": msg})
                return df

        # read from S3
        dfs = []
        for path in paths:
            _df = self.read_s3_data(path, params)
            if _df is None:
                msg += f"Path error, skipping: {path}" + "\n"
                continue
            msg += f"Read from path: {path}" + "\n"
            dfs.append(_df)
        df = pd.concat(dfs)

        logger.debug(f"Read {dataset}", extra={"transform": self.name})

        # Cache it for future use
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if (len(paths) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": paths,
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec['name'] + "-raw", f"Raw Dataset: {spec['name']}", df, lineage)

        return df    

    def get_dataset_db(self, spec):
        name = spec['name']
        source = spec['source']

        testmode = self.args.get('testmode', False)
        env      = 'test' if testmode else 'prod'

        # Get the db engine
        engine_name = source['params'][env]['engine']
        engine = self.engines[engine_name]

        # Get the query
        query = source['params'][env]['query']
        if query.endswith('.sql'):
            # we have a file containing the query
            sqlfile = os.path.join(self.scriptdir, query)
            query = open(sqlfile).read()

        df = self.read_db_source(engine, query)

        # Now the input load...
        lineage = {
            "type": "lineage",
            "transform": self.name,
            "dependencies": [
                {
                    "type": "database",
                    "nature": "input",
                    "objects": [source['dataset']],
                },
            ],
        }

        self.update_frame(spec, f"Dataset: {name}", df, lineage)

        return df

    def read_db_source(self, engine, query):
        # Run the query
        df = pd.read_sql_query(satext(query), con=engine)

        return df

    def get_datewindow(self, source, spec):
        datewindow = {}
        default_delta = 180

        msg = ""
        start_date_str = source.get('start_date')
        end_date_str = source.get('end_date')

        if not end_date_str:
            msg += f"End date not specified. Using default as yesterday's date\n"
            end  = self.args['run_date'] # yesterday
            end_date_str = end.isoformat()

        if not start_date_str:
            msg += f"start date not specified. Using default as 180 days prior to end date\n"
            start = dateparser.parse(end_date_str).date() + timedelta(days=-default_delta)
            start_date_str = start.isoformat()

       # If start and end dates are specified, check if date is parseable
        try:
            start_date = dateparser.parse(start_date_str).date()
            end_date = dateparser.parse(end_date_str).date()
        except:
            logger.exception(
                        f"Invalid start date or end date. Skipping {spec['name']}.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4, cls=SafeEncoder)}
                    )
            datewindow = None # be explicit
            return datewindow

        if start_date > end_date:
            logger.exception(
                        f"Start date > end date. Skipping {spec['name']}.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4, cls=SafeEncoder)}
                    )
            datewindow = None # be explicit
            return datewindow

        datewindow['start_date'] = start_date
        datewindow['end_date'] = end_date

        logger.debug(
            f"{spec['name']}: {datewindow['start_date']} to {datewindow['end_date']}",
            extra={"transform": self.name, "data": json.dumps(datewindow, indent=4, cls=SafeEncoder)}
        )        
        return datewindow
    
    def load_dataset(self, spec, source, datewindow):
        msg = ""

        source_id       = source.get('id', source.get('source_id', None))
        dataset         = source['dataset']
        source_version  = source.get('source_version', 'v1')
        nature          = source.get('nature', 's3')

        start_date      = datewindow['start_date']
        end_date        = datewindow['end_date']

        if source_id == "custom":
            # we have a custom defined method in the derived class to generate the dataset
            if not hasattr(self, dataset):
                logger.exception(
                    f"No handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4, cls=SafeEncoder)}
                )
                return None

            handler = getattr(self, dataset)
            params = source.get("params", {})
            df = handler(start_date, end_date, params)
            if df is None:
                logger.exception(
                    f"Couldn't load dataset for {spec['name']}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4, cls=SafeEncoder)}
                )
                return None

            msg += f"Loaded using custom method: {dataset}" + "\n"
        else:

            if source_id is None and dataset is None:
                logger.error(
                    f"source id and name missing",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4, cls=SafeEncoder)}
                )                
                return None

            # we are using the SDK to get the dataset
            datacred = self.args['datacred']
            doodle = TransformDoodle(self, self.state, datacred)

            # Now get the source details...
            d_source, paths = doodle.get_source_paths(start=start_date,
                                                      end=end_date,
                                                      name=dataset,
                                                      version=source_version,
                                                      source_id=source_id)

            msg += f"Paths: {paths}" + "\n"

            # Insert a read action..
            try:
                result = doodle.update_source(d_source['id'], {})
                logger.debug(f"Updated Doodle: {dataset}",
                             extra={
                                 'transform': self.name,
                                 'data': json.dumps(result, indent=4, cls=SafeEncoder)
                             })
            except:
                logger.exception(f"Unable to update Doodle {dataset}",
                                 extra={
                                     'transform': self.name
                                 })

            if nature == "s3":
                # read the data from S3
                df = self.get_dataset_s3(spec, source, paths, start_date, end_date)
            else:
                logger.error(f"Unable to access raw data",
                             extra={
                                 'transform': self.name
                             })                


        # check if we need to perform a post-processing step on the input data
        postprocess = source.get("postprocess")
        if postprocess != None:
            if not hasattr(self, postprocess['method']):
                logger.exception(
                    f"No post-process handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4, cls=SafeEncoder)}
                )
            else:
                handler = getattr(self, postprocess['method'])
                params = postprocess.get("params", {})
                df = handler(df, params)
                msg += f"Post-processed dataset={dataset} using handler: {postprocess}" + "\n"

        msg += note(df, f"Input Dataset: {dataset}")
        logger.debug(
            f"Loaded dataset {dataset}",
            extra={"transform": self.name, "data": msg}
        )

        return df    

    def s3_store_result(self, spec, result, extra):
        name        = spec['name']
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = time.time()


        # where are we storing it?
        targetdir = os.path.join(self.args['s3root'], f"{name}/{run_date}/{epoch}")

        resultfile = os.path.join(targetdir, f"data.csv")
        metadatafile = os.path.join(targetdir, f"metadata.json")

        metadata = self.get_default_metadata(self.state)
        metadata['spec'] = spec
        metadata.update(extra)
        
        # write to s3
        s = json.dumps(result, indent=4, cls=SafeEncoder)
        with s3.open(resultfile, 'w') as fd:
            result.to_csv(fd, index=False)
        with s3.open(metadatafile, 'w') as fd:
            json.dump(metadata, fd, indent=4, cls=SafeEncoder)

        msg = f"s3 location: {resultfile}" + "\n"
        msg += f"metadata location: {metadatafile}" + "\n"

        logger.debug(f"Wrote final results to S3: {name}",
                        extra={"transform": self.name,
                                "data": msg})

    def db_store_result(self, spec, result, extra):
        name    = spec['name']
        store   = spec['store']

        testmode = self.args.get('testmode', False)
        env      = 'test' if testmode else 'prod'

        # Get the db engine
        engine_name = store['params'][env]['engine']
        engine = self.engines[engine_name]

        # write results to db
        table_name = store['params'][env]['table']
        ret = results.to_sql(table_name,
                            engine,
                            if_exists='append',
                            index=False)
        # write data to db
        data_table_name = name.replace('-','_')
        ret = data.to_sql(data_table_name,
                            engine,
                            if_exists='append',
                            index=False)

        msg = f"DB engine: {engine}" + "\n"

        logger.debug(f"Wrote monitor results to DB",
                        extra={"transform": self.name,
                                "data": msg})

    def store_result(self, spec, result, extra):
        name    = spec['name']
        store   = spec.get('store', {"sink": "s3"})

        for f in ["sink"]:
            if f not in store:
                logger.exception(
                    f"Store has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4, cls=SafeEncoder)}
                )
                return

        sink = store['sink']
        if sink == "s3":
            # store in s3
            self.s3_store_result(spec, result, extra)
        elif sink == "db":
            # store in db
            self.db_store_result(spec, result, extra)
        else:
            logger.exception(f"Unknown store for dataset: {name}",
                         extra={
                             'transform': self.name
                         })

    ###########################################
    # Implement the data transformation logic...
    ###########################################
    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra={"transform": self.name}
        )

        # Will be used in other places..
        self.state = state

        # Get the profile spec
        is_valid, profile, msg = profilespec.get_profile(self, "policyapp.dataqualityv2")
        if is_valid:
            name = profile.get('name', 'unknown')
            logger.debug(
                f"Loaded profilespec: {name}",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")

        specs = profile.get("specs", None)
        if specs is None:
            raise Exception("Could not find 'specs' in profile")

        # Now go through each spec and process it
        for spec in specs:

            ## first, some checks on the spec
            do_process_spec = True
            name = spec.get('name', 'NO_SPEC_NAME')

            enabled = spec.get("active", True)
            if not enabled:
                logger.debug(
                    f"Spec not enabled, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4, cls=SafeEncoder)}
                )
                do_process_spec = False
                continue

            for f in ["name", "source", "transformations"]:
                if f not in spec:
                    logger.exception(
                        f"Spec has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4, cls=SafeEncoder)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            ## we can now proceed with processing the spec            
            source = spec['source']
            
            # get time_window for indicator and observation data set
            datewindow = self.get_datewindow(source, spec)
            if datewindow is None :
                do_process_spec = False
                continue
            
            ## we can now proceed with processing the spec
            # first, load the source data
            data = self.load_dataset(spec, source, datewindow)

            # then, process it
            result, msg = self.process_spec(spec, data)
            if result is None:
                continue

            ## store the expectation validation result
            self.store_result(spec, result, {'notes': msg})


        # Done
        logger.debug(
            "Complete execution", extra={"transform": self.name}
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



from enrichsdk.app.widget.basewidget import BaseWidget
from enum import Enum

from enrichsdk.app.widget.customcomponent import BodyComponent, HeaderComponent


class CustomWidget():
    widgets = []

    def add_widget(self, widget):
        if widget:
            self.widgets.append(widget)

    def get_widgets(self):
        return self.widgets

    """ "name": f"Samples from {table_details.get('table')}",
                "description": persona.get('description'),
                "type": rendering_view,
                "results_count": widget_results,
                "structure": "flat",
                "columns": columns,
                "rows": [
                    v
                    for v in rows
                ],
                "search": search,
                "search_columns": search_columns,
                "body_components": body_components,
                "header_components": header_components,
                "td_class": "td-align-left td-width-200 wordwrap" """


class TableStructureType(str, Enum):
    FLAT = 'flat'
    S3 = 'S3'


class TableWidget(BaseWidget):
    def __init__(self, widgetid, name, title, template, description, result_count, table_structure, columns, rows):
        super().__init__(widgetid, name, title, template)
        self.rows = rows
        self.css_class = ''
        self.columns = columns
        self.type = table_structure
        self.result_count = result_count
        self.description = description
        self.search = None

    @property
    def css_class_attr(self):
        return self.css_class

    @css_class_attr.setter
    def css_class_attr(self, value):
        self.css_class = value

    def search_attr(self):
        return self.search

    def search_attr(self, value):
        self.search = value

    # @property
    # def search_columns_prop(self):
    #     return self.search_columns
    #
    # @search_columns_prop.setter
    # def search_columns_prop(self, value):
    #     self.search_columns = value

    """ "url": f"?persona={persona_name}&table={table}&query={querystr}&process=download",
        "id": "download",
        "label": "",
        "title": "Download All Matched Results",
        "alt": "",
        "class": "",
        "template": "action_icon",
        "icon": "download-blue-icon" """


class TextWidget(BaseWidget):
    def __init__(self, widgetid, name, title, template):
        super().__init__(widgetid, name, title, template)
        self.css_class = ''
        self.header_components = None
        self.body_components = None

    @property
    def css_class_attr(self):
        return self.css_class

    @css_class_attr.setter
    def css_class_attr(self, value):
        self.css_class = value

    def get_header_components(self):
        return self.header_components

    def set_header_components(self, value):
        if isinstance(value, HeaderComponent):
            self.header_components = value.get_components()

    def get_body_components(self):
        return self.body_components

    def set_body_components(self, value):
        if isinstance(value, BodyComponent):
            self.body_components = value.get_components()



class ActionItemWidget(BaseWidget):
    def __init__(self, widgetid, name, title, template):
        super().__init__(widgetid, name, title, template)



#!/usr/bin/env python3


# Core
import os
import sys
import pip
import base64
import subprocess
import textwrap
import json
import imp
import yaml
import traceback
import pkg_resources
import logging, traceback
import pyfiglet
import warnings

# thirdparty
import click
from pythonjsonlogger import jsonlogger
from prompt_toolkit import prompt

# Disable warnings
import urllib3
urllib3.disable_warnings()

# self
import enrichsdk
from enrichsdk import (
    package,
    datasets,
    api as sdkapi,
    lib,
    featurestore as fslib,
    utils,
)
from enrichsdk.package.validators import GenericDescValidator

class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):

        if cmd_name == 'bootstrap':
            cmd_name = 'init'

        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        print(self.list_commands(ctx))
        print(cmd_name)
        matches = [x for x in self.list_commands(ctx)  if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


def get_backend_obj(ctx, backend):
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    backends = config["backends"]

    relevant = None
    for b in backends:
        if (b.get("server", None) == backend) or (b.get("name", None) == backend):
            relevant = b
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))

    # Decode the key before accessing
    for col in ['key', 'htpasswd_user', 'htpasswd_pwd']:
        value = base64.b64decode(relevant[col].encode("utf-8")).decode("utf-8")
        relevant[col] = value

    backend = sdkapi.Backend(relevant)
    return backend

@click.group(cls=AliasedGroup)
def process():
    """
    init/test/install Enrich modules and access server

    \b
    Getting started:
       version:  Version of this sdk
       start:    First time instructions
       env:      Setup/check the setup

    \b
    Development:
       init:   Bootstrap modules including transforms*
       test:   Test transforms, manage datasets
       doodle: Access Doodle metadata server
       manage: Manage services such as mongo

    \b
    Server:
       api:       Access the server API

    \b
    Utils:
       sample:    Sample data for sharing

    \b
    Helpers:
       show-log:  Pretty print log output

    *Command used to be called bootstrap
    """
    pass


@click.command("version", hidden=True)
def _version():
    """
    Show version
    """
    print(enrichsdk.__version__)


@click.command("start", hidden=True)
@click.option("--minimal/-m", default=False)
def _start(minimal):
    """
    Instructions to get started
    """

    result = pyfiglet.figlet_format("Enrich SDK", font="slant")
    print(result)

    home = os.environ["HOME"]
    if "ENRICH_ROOT" in os.environ:
        enrich_root = os.environ["ENRICH_ROOT"]
    else:
        enrich_root = os.getcwd() + "/enrich"

    if "VIRTUAL_ENV" in os.environ:
        venv_root = os.environ["VIRTUAL_ENV"]
        venv_name = os.path.abspath(os.path.basename(venv_root))
        install_settings = [
            'echo "source {}/bin/activate" > $ENRICH_ROOT/env.sh'.format(venv_root),
            'echo "#workon {}" >> $ENRICH_ROOT/env.sh'.format(venv_name),
        ]
    elif "CONDA_PREFIX" in os.environ:
        venv_root = os.environ["CONDA_PREFIX"]
        venv_name = os.path.basename(venv_root)
        install_settings = [
            'echo "conda activate {}" >> $ENRICH_ROOT/env.sh'.format(venv_name),
        ]
    else:
        print("Please use virtualenvwrapper or conda to install sdk")
        return

    steps = [
        {
            "description": "Understand Enrich and SDK",
        },
        {
            "description": "set ENRICH_ROOT and populate. This will create the directory structure required to make enrich work. You will see several directories that are required at various points in the lifecycle of the project",
            "cmd": [
                f"export ENRICH_ROOT={enrich_root}",
                "enrichpkg env populate",
            ],
        },
        {
            "description": "Check and update siteconf and versionmap. Siteconf stores site-specific configuration required by pipelines including credentials, security policy etc.. Versionmap (which you dont have to worry about right now) allows the pipelines to track the version information of all the modules deployed",
            "cmd": [
                "enrichpkg env check",
                "cat $ENRICH_ROOT/etc/siteconf.json",
            ],
        },
        {
            "description": "[OPTIONAL] For complex deployments, create an environment/context file. Talk to Scribble before you use this.",
            "cmd": "enrichpkg env sample-context > context.yaml",
        },
        {
            "description": "[OPTIONAl] Create a simple settings file. This is useful in everyday execution. You can simply source this script and then run the enrichpkg",
            "cmd": [
                "# file to be sourced before you start working. use the appropriate",
                "# environment activation mechanism. Check the script generated by",
                "# populate",
                "",
                "cat $ENRICH_ROOT/env.sh",
                'echo "source $ENRICH_ROOT/env.sh" >> $HOME/.bashrc',
                "",
                "# If the file doesnt exist or doesnt look right, use the following"
            ]
            + install_settings
            + [
                'echo "export ENRICH_ROOT={}" >> $ENRICH_ROOT/env.sh'.format(
                    enrich_root
                )
            ],
        },
        {
            "description": "Customer directory has actual code repositories. Your code will be stored here. Change dir to $ENRICH_CUSTOMERS. You will find that it is empty but dont panic",
            "cmd": "cd $ENRICH_ROOT/customers",
        },
        {
            "cmd": [
                "#To handle python paths etc. avoid spaces and hyphen in the names",
                "git clone https://github.com/pingali/enrich-acme.git acme",
            ],
            "description": "GIT checkout an example code repository. This time it is from a company called acme. You will have a similar repository and checkout. The process is similar whether it is on the local machine or on the server",
        },
        {
            "cmd": "cd $ENRICH_ROOT/customers/acme",
            "description": "Change to checked out repository",
        },
        {
            "cmd": [
                "# if acme is already populated",
                "enrichpkg init repo -p ."
            ],
            "description": "[OPTIONAL] Initialize the repo, if not already done. In case of Acme you dont have tEnrich requires a particular structure to the repository. You will notice __init__.py and enrich.json. These are files that tell the Enrich system what is present where.",
        },
        {
            "cmd": [
                "# Typically usecases have first letter in capital",
                "enrichpkg init usecase -p Marketing",
            ],
            "description": "[OPTIONAL] The code is organized in terms of projects or usecase groups. Create one usecase, say Marketing. In case of acme, you dont have to do this. Just look around."
        },
        {
            "cmd": [
                "cd Marketing"
            ],
            "description": "Change to usecase"
        },
        {
            "description": "Initialize a simple transform. Transform is a python module. It could be a single file, directory or a full package.",
            "cmd": [
                "# Will create an example hello world script",
                "enrichpkg init transform-helloworld -p transforms/helloworld.py",
            ]
        },
        {
            "description": "[OPTIONAL] You could do more complex transforms. You can initialize a python module directory or a full package. You can come back to this later on",
            "cmd": [
                "# Will create a more simple python script that is more complex than helloworld",
                "enrichpkg init transform-simple -p transforms/simplefile.py",
                "",
                "# Will create a python module (a full directory)",
                "enrichpkg init transform-simple -p transforms/simpledir",
                "",
                "# Will create a python package. Dont use this unless you are sure",
                "enrichpkg init transform-package -p transforms/simplepackage",
            ],
        },
        {
            "cmd": [
                "# Add any requirements and install them",
                "cat ../requirements.txt",
                "pip install -r ../requirements.txt",
            ],
            "description": "[OPTIONAL] Each repository has additional requirements going beyond what enrich requires, e.g., a module to connect to snowfake. Install these repo-specific requirements to enable the transform to run. You can come back to this later on",
        },
        {
            "cmd": [
                "vi transforms/helloworld.py",
            ],
            "description": "Eyeball the helloworld transform",
        },
        {
            "cmd": [
                "enrichpkg test transform transforms/helloworld.py",
            ],
            "description": "You have to locally test the transform to make sure everything is working as planned including the parameters"
        },
        {
            "cmd": [
                "enrichpkg init pipeline -p pipelines/conf/helloworld.py"
            ],
            "description": "Transforms are like plugins. They have to be incorporated into a pipeline to ensure that it runs on server. Initialize a pipeline. Note the path to the pipelines. They have a particular structure as described in the SDK and at particular location. The Enrich execution engine searches these paths",
        },
        {
            "cmd": ["enrichpkg test conf --capture pipelines/conf/helloworld.py"],
            "description": "Sanity check the pipeline to ensure that everything is as per plan. Note that in some cases where credentials are being accessed etc. we will have additional setup"
        },
        {
            "cmd": [
                "# prefect is the default workflow engine",
                "enrichpkg init prefectjob -p workflows/prefect/daily.py",
            ],
            "description": "A workflow allows us to combine multiple pipelines and multiple runs of each into a single activity. We use prefect for the same. We could use something else. We dont need prefect cloud though. We will run the workflow from our scheduler"
        },
        {
            "cmd": [
                "cat workflows/prefect/daily.py",
                "python3 workflows/prefect/daily.py"
            ],
            "description": "Check the content of the workflow. Once ready you can include this workflow in the nextdoor scheduler"
        },
    ]

    for i, s in enumerate(steps):
        print("[{:2}] {}".format(i + 1, "\n".join(textwrap.wrap(s["description"]))))
        print("")
        if minimal:
            continue
        if "cmd" in s:
            if isinstance(s["cmd"], str):
                cmds = [s["cmd"]]
            else:
                cmds  = s["cmd"]
            for c in cmds:
                if not c.startswith("#") and len(c) > 0:
                    print("    $ {}".format(c))
                else:
                    print("    {}".format(c))
        print("")


@click.command("init", hidden=True)
@click.argument("component")
@click.option(
    "--path", "-p", required=False, default=None, help="Directory where the files should be stored"
)
@click.option(
    "--template",
    "-t",
    required=False,
    default=None,
    help="Template to override the defaults",
)
@click.option("--context", required=False, default=None, help="Environment file")
def init_(component, path, template, context):
    """
    Bootstrap/initialize a fresh module

    Modules currently supported include transform, pipeline and asset. Use
    help to list possible components::

         enrichpkg init help*

    For now only the transforms and models are testable locally
    without the need for a test server.

    In future we expect to more comprehensively support the
    development of all modules.

    A custom template can be provided as well.

    * init command used to be called bootstrap

    """

    components = [
        ["repo" , "A new git repository"],
        ["usecase", "A collection of transforms, pipelines, assets and workflows"],
        ["transform-package", "A complex transform that is structured as a full python package"],
        ["transform-simple", "A transform structured as a python module"],
        ["transform-helloworld", "A simple hello world transform for study"],
        ["transform-query", "A  transform that uses a pre-existing database querying module"],
        ["transform-metrics", "A  transform that uses a pre-existing metrics module"],
        ["transform-iris", "Simple Iris post-processing example"],
        ["asset", "A reusable library "],
        ["dashboard", "A django app that fits the Enrich Dashboard framework"],
        ["pipeline", "A pipeline configuration"],
        ["prefectjob", "Prefect workflow template"],
        ["app-index", "Index page of a new app"],
        ["datasets", "Datasets specification file"],
        ["rscript", "R-Script template"],
        ["pyscript", "A python script template"],
        ["singlepageapp", "Single page app"]
    ]

    if component not in [c[0] for c in components]:
        print("The following components are supported:\n")
        for c in components:
            print(f"   {c[0]:20}: {c[1]}")
        print("\nUse the component name like this")
        print("    enrichpkg init repo -p <path>")
        print("")
        print("You can define your own templates as well. Please use the -t option\n")
        return

    if path is None:
        print("Path should specified using -p option")
        return

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise
    context.set_env()

    if not package.checkenv(validate_only=True):
        print("Environment is incomplete")
        print("Try: enrichpkg env check")
        return

    package.bootstrap(component, path, template)


@click.command("show-log", hidden=True)
@click.argument("logfile")
def _show_log(logfile):
    """
    Pretty-print run log
    """

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.print_log(logfile)


@click.group("env", hidden=True)
@click.pass_context
def envsetup(ctx):
    """
    Setup the environment. Environment includes:\n
      \b
      (a) Workspace for enrich to work
      (b) Minimal configuration

    This can be specified using environment variables
    and/or 'context' file.

    The absolute minimal required environment variable
    is ENRICH_ROOT (path to enrich workspace e.g.,
    ~/enrich)
    """

    ctx.ensure_object(dict)


@envsetup.command("sample-context")
@click.option("--root", "-r", default=None, help="Root of enrich")
def _context_generate(root):
    """
    Generate sample context file
    """

    if root is not None:
        os.environ["ENRICH_ROOT"] = os.path.abspath(root)
    elif "ENRICH_ROOT" not in os.environ:
        os.environ["ENRICH_ROOT"] = os.path.expandvars("$HOME/enrich")

    context = lib.Context()
    context = context.asdict()

    if "siteconf" not in context:
        context["siteconf"] = os.path.join(context["ENRICH_ETC"], "siteconf.json")

    print("Please store this in context.yaml")
    print("Minimum required valid paths include: siteconf, versionmap, enrich_data")
    print("Example: enrichpkg api --context context.yaml ...")
    print("----")
    print(yaml.dump(context, indent=4))


@envsetup.command("sample-versionmap")
def _versionmap():
    """
    Generate sample versionmap
    """

    print(
        "Please store this in versionmap.json. Default is $ENRICH_ROOT/etc/versionmap.json"
    )
    print("Update context.yaml with this versionmap path")
    print("----")
    print(package.get_sample_versionmap())


@envsetup.command("dump-versionmap")
def _dump_versionmap():
    """
    Show versionmap
    """

    timestamp, versionmap = lib.read_versionmap()
    print(json.dumps(versionmap, indent=4))

@envsetup.command("sample-siteconf")
def _siteconf():
    """
    Generate sample siteconf
    """

    print(
        "Please store this in siteconf.json. Default is $ENRICH_ROOT/etc/siteconf.json"
    )
    print("Update context.yaml with this siteconf path")
    print("----")
    print(package.get_sample_siteconf())

@envsetup.command("sample-djangoenv")
def _env():
    """
    Sample django-env environment variables.
    """

    print(
        "Please store this in django-env. Default is $ENRICH_ROOT/etc/django-env"
    )
    print(package.get_sample_djangoenv())


@envsetup.command("show-credential")
@click.argument("name")
def _show_cred(name):
    """
    Generate sample siteconf
    """
    print(json.dumps(lib.get_credentials_by_name(name)))


@envsetup.command("populate")
@click.option("--context", default=None, help="Context file")
def _populate(context):
    """
    Populate the directories
    """
    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        print("Error! " + str(e))
        return

    context.set_env()

    package.populate(context=context.asdict())


@envsetup.command("check")
@click.option("--context", default=None, help="Context file")
def _checkenv(context):
    """
    Check environment
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        print("Error! " + str(e))
        return

    context.set_env()

    package.checkenv(context=context)


##########################################################
# Test command
##########################################################
@click.group("test", hidden=True)
@click.option("--context", default=None, help="Environment file")
@click.pass_context
def test(ctx, context):
    """
    Test transforms/pipelines etc.
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise

    context.set_env()
    ctx.ensure_object(dict)
    ctx.obj["context"] = context.asdict()


@test.command("transform")
@click.argument("pkgdir")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def test_pkg(ctx, pkgdir, capture):
    """
    Unit testing of a package module (transform)
    """
    context = ctx.obj["context"]

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg env check")
        raise Exception("Environment incomplete")

    package.test_transform(pkgdir, capture, context)


@test.command("task-lib")
@click.argument("taskdir")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def test_task(ctx, taskdir, capture):
    """
    Unit testing of a task library
    """

    context = ctx.obj["context"]

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg env check")
        raise Exception("Environment incomplete")

    package.test_task(taskdir, capture, context)


@test.command("conf")
@click.argument("spec")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def test_conf(ctx, spec, capture):
    """
    Minimal testing of pipeline/task configuration
    """

    context = ctx.obj["context"]

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg env check")
        raise Exception("Environment incomplete")

    package.test_conf(spec, capture, context)


@click.group("data")
@click.argument("spec")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def testdata(ctx, spec, capture):
    """
    Manage test data


    spec could be transform or a spec file
    """
    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg env check")
        raise Exception("Environment incomplete")

    ctx.ensure_object(dict)
    ctx.obj["spec"] = spec
    ctx.obj["capture"] = capture


@testdata.command("list")
@click.pass_context
def _testdata_list(ctx):
    """
    List available test datasets
    """

    package.testdata_action(
        "list", ctx.obj["spec"], ctx.obj["capture"], ctx.obj["context"]
    )


@testdata.command("show")
@click.argument("dataset")
@click.pass_context
def _testdata_show(ctx, dataset):
    """
    Show the details of a given dataset
    """

    package.testdata_action(
        "show", ctx.obj["spec"], ctx.obj["capture"], ctx.obj["context"], name=dataset
    )


@testdata.command(
    "run",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("dataset")
@click.argument("command")
@click.pass_context
def _testdata_action(ctx, dataset, command):
    """
    Generate action commands for a dataset over a specified range

    Pass additional parameters as key-value pairs as
    shown in show.
    """

    args = dict()
    for item in ctx.args:
        args.update([item.split("=")])

    package.testdata_action(
        "action",
        ctx.obj["spec"],
        ctx.obj["capture"],
        ctx.obj["context"],
        name=dataset,
        command=command,
        **args
    )


test.add_command(testdata)

##########################################################
# Doodle
##########################################################
@click.group("doodle", hidden=True)
@click.option("--context", default=None, help="Environment file")
@click.option("--cred", default="doodle", help="Doodle credential name")
@click.pass_context
def doodle(ctx, context, cred):
    """
    Access doodle server
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise

    context.set_env()
    ctx.ensure_object(dict)
    ctx.obj["context"] = context.asdict()
    ctx.obj["cred"] = cred

@doodle.command("list-catalogs")
@click.pass_context
def _list_catalogs(ctx):
    """
    List catalogs
    """

    cred = enrichsdk.lib.get_credentials_by_name(ctx.obj['cred'])
    doodle = datasets.Doodle(cred)
    catalogs = doodle.list_catalogs()
    print(json.dumps(catalogs, indent=4))

@doodle.command("search-catalogs")
@click.argument("name")
@click.option("-v", "--version", default="v1")
@click.pass_context
def _search_catalogs(ctx, name, version):
    """
    Search catalogs
    """

    cred = enrichsdk.lib.get_credentials_by_name(ctx.obj['cred'])
    doodle = datasets.Doodle(cred)
    catalogs = doodle.search_catalogs(name=name,version=version)
    print(json.dumps(catalogs, indent=4))

@doodle.command("list-sources")
@click.pass_context
def _list_sources(ctx):
    """
    List sources
    """

    cred = enrichsdk.lib.get_credentials_by_name(ctx.obj['cred'])
    doodle = datasets.Doodle(cred)
    sources = doodle.list_sources()
    print(json.dumps(sources, indent=4))

@doodle.command("search-sources")
@click.argument("name")
@click.option("-v", "--version", default="v1")
@click.pass_context
def _search_sources(ctx, name, version):
    """
    Search sources
    """

    cred = enrichsdk.lib.get_credentials_by_name(ctx.obj['cred'])
    doodle = datasets.Doodle(cred)
    sources = doodle.search_sources(name=name,version=version)
    print(json.dumps(sources, indent=4))

@doodle.command("show-source")
@click.argument('source_id')
@click.pass_context
def _show_source(ctx, source_id):
    """
    Show source
    """

    cred = enrichsdk.lib.get_credentials_by_name(ctx.obj['cred'])
    doodle = datasets.Doodle(cred)
    source = doodle.get_source(source_id)
    print(json.dumps(source, indent=4))


@doodle.command("update-source")
@click.argument('source_id')
@click.argument('filename')
@click.pass_context
def _update_source_metadata(ctx, source_id, filename):
    """
    List sources
    """

    cred = enrichsdk.lib.get_credentials_by_name(ctx.obj['cred'])
    doodle = datasets.Doodle(cred)

    try:
        update = json.load(open(filename))
    except Exception as e:
        print("Error!", e)
        print("Unable to read filename. It should be a valid json file")

    result = doodle.update_source(source_id, update)
    print(json.dumps(result, indent=4))

#########################################
# API
#########################################
@click.group(hidden=True)
@click.option("--config", default=None, help="API Config File")
@click.pass_context
def api(ctx, config):
    """
    Access Enrich Server
    """

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg env check")
        raise Exception("Environment incomplete")

    alt = os.path.expandvars("$ENRICH_ETC/api.json")
    if config is None:
        config = alt

    # Write a default
    default = {"schema": "v1:api:enrich", "backends": []}
    if not os.path.exists(config):
        with open(config, "w") as fd:
            fd.write(json.dumps(default, indent=4))

    ctx.ensure_object(dict)
    ctx.obj["config"] = config


####################################
# Support for sampling data
####################################
@click.group("sample", hidden=True)
@click.pass_context
def sample(ctx):
    """
    Sample data from blockstores
    """
    pass


@sample.command("crawl")
@click.argument("root")
@click.option("--ext", default="json", help="Extension of files to look for")
@click.pass_context
def crawl(ctx, root, ext):
    """
    Crawl the root to find possible files

    Args:
       root: Path in blockstore with s3:// prefix
    """

    utils.sample_crawl(root, ext)


@sample.command("cat")
@click.argument("path")
@click.pass_context
def crawl(ctx, path):
    """
    Show the contents of path

    Args:
       path: Path to file in blockstore with s3:// prefix
    """

    utils.sample_cat(path)


####################################
# Configure access to server
####################################
@click.group()
@click.pass_context
def config(ctx):
    """
    Configure access to Enrich Server
    """
    pass


@config.command("init")
@click.pass_context
def _config_init(ctx):
    """
    Initialize the config file
    """
    default = {
        "schema": "v1:api:enrich",
        "backends": [],
        "services": [],
    }
    filename = ctx.obj["config"]
    if os.path.exists(filename):
        print("API access configuration file already exists. Please remove it first")
        print("See:", filename)
        return

    with open(filename, "w") as fd:
        fd.write(json.dumps(default, indent=4))

    print("Initialized config")


@config.command("list")
@click.pass_context
def _config_list(ctx):
    """
    List available backends/services
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    print("Backends")
    print("-----")
    backends = config["backends"]
    for b in backends:
        key = base64.b64decode(b["key"].encode("utf-8")).decode("utf-8")
        try:
            print("{} [{} @ {}]".format(b["name"], key, b["server"]))
        except:
            pass

    print("\n")
    print("Services")
    print("-----")
    services = config.get("services", [])
    for s in services:
        try:
            print("{} [{}]".format(s["name"], s["path"]))
        except:
            pass


@config.command("backend-add")
@click.argument("name")
@click.argument("server")
@click.pass_context
def _backend_add(ctx, name, server):
    """
    Add backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    key = prompt(">API Key: ")
    htpasswd_user = prompt(">HTPasswd Username: ")
    htpasswd_pwd = prompt(">HTPasswd Password: ",)

    if len(key) == 0 or len(htpasswd_user) == 0 or len(htpasswd_pwd) == 0:
        print("Empty APIKey or HTPasswd username or password")
        return

    # Encode the key before storing
    key = base64.b64encode(key.encode("utf-8")).decode("utf-8")
    htpasswd_user  = base64.b64encode(htpasswd_user.encode("utf-8")).decode("utf-8")
    htpasswd_pwd  = base64.b64encode(htpasswd_pwd.encode("utf-8")).decode("utf-8")

    backends = config["backends"]
    missing = True
    for b in backends:
        if ("server" in b) and (b["server"] == server):
            b["name"] = name
            b["server"] = server
            b["key"] = key
            b["htpasswd_user"] = htpasswd_user
            b["htpasswd_pwd"] = htpasswd_pwd
            missing = False
            break

    if missing:
        backends.append({
            "name": name,
            "server": server,
            "key": key,
            "htpasswd_user": htpasswd_user,
            "htpasswd_pwd": htpasswd_pwd,
        })

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Added {}".format(name))


@config.command("backend-remove")
@click.argument("name")
@click.pass_context
def _backend_remove(ctx, name):
    """
    Add backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    backends = config["backends"]
    missing = True
    updated = [
        b
        for b in backends
        if ((b.get("server", None) != name) and (b.get("name", None) != name))
    ]

    if len(backends) == len(updated):
        raise Exception("Backend not found: {}".format(name))

    config["backends"] = updated

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Removed {}".format(name))


@config.command("service-add")
@click.argument("name")
@click.argument("path")
@click.pass_context
def _service_add(ctx, name, path):
    """
    Add URL
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    if "services" not in config:
        config["services"] = []

    services = config["services"]
    missing = True
    for s in services:
        if ("name" in s) and (s["name"] == name):
            s["name"] = name
            s["path"] = path
            missing = False
            break

    if missing:
        services.append({"name": name, "path": path})

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Added {}".format(name))


@config.command("service-remove")
@click.argument("name")
@click.pass_context
def _service_remove(ctx, name):
    """
    Remove service
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    services = config["services"]
    missing = True
    updated = [s for s in services if s.get("name", None) != name]

    if len(services) == len(updated):
        raise Exception("Service not found: {}".format(name))

    config["services"] = updated

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Removed {}".format(name))


@click.group()
@click.argument("backend")
@click.pass_context
def show(ctx, backend):
    """
    Access a backend
    """
    ctx.obj["backend"] = get_backend_obj(ctx, backend)


@click.command("health")
@click.argument("backend")
@click.pass_context
def _health(ctx, backend):
    """
    Show system health
    """

    backend = get_backend_obj(ctx, backend)
    result = backend.health()

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    sdkapi.draw_dict(backend, "Health", result)


@click.command("app-list")
@click.argument("backend")
@click.pass_context
def _app_list(ctx, backend):
    """
    Show apps available
    """

    backend = get_backend_obj(ctx, backend)
    try:
        result = backend.app_list()
        print(json.dumps(result, indent=4))
    except:
        traceback.print_exc()

@click.command("app-detail")
@click.argument("backend")
@click.argument("name")
@click.pass_context
def _app_detail(ctx, backend,name):
    """
    Show detail of an app
    """

    backend = get_backend_obj(ctx, backend)
    try:
        result = backend.app_detail(name)
        print(json.dumps(result, indent=4))
    except:
        traceback.print_exc()


@click.command("pipeline-runs")
@click.argument("backend")
@click.argument("usecase")
@click.argument("pipeline")
@click.pass_context
def _run_list(ctx, backend, usecase, pipeline):
    """
    Show available runs for pipeline
    """

    backend = get_backend_obj(ctx, backend)
    result = backend.run_list(usecase, pipeline)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        if "message" in result:
            print("Reason:", result["message"])
        return

    result["data"] = sorted(result["data"], key=lambda r: r["start_time"], reverse=True)

    print(json.dumps(result['data'], indent=4))
    #columns = ["runid", "status", "start_time", "end_time"]
    #sdkapi.draw(
    #    backend, "run-list", result['data'], columns, {"Usecase": usecase, "Pipeline": pipeline}
    #)

@click.group("scheduler")
@click.argument("backend")
@click.pass_context
def scheduler(ctx, backend):
    """
    Access scheduler at backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    backends = config["backends"]

    relevant = None
    for b in backends:
        if (b.get("server", None) == backend) or (b.get("name", None) == backend):
            relevant = b
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))

    # Decode the key before accessing
    for col in ['key', 'htpasswd_user', 'htpasswd_pwd']:
        value = base64.b64decode(relevant[col].encode("utf-8")).decode("utf-8")
        relevant[col] = value

    backend = sdkapi.Backend(relevant)
    ctx.obj["backend"] = backend


@scheduler.command("list")
@click.pass_context
def _scheduler_list(ctx):
    """
    Show the scheduled jobs
    """
    backend = ctx.obj['backend']
    try:
        result = backend.scheduler_jobs_list()

        jobs = result['jobs']

        #print(json.dumps(jobs, indent=4))
        for j in jobs:
            schedule = "%(minute)s %(hour)s %(day)s %(month)s %(day_of_week)s" % j
            j['cron'] = schedule
            for col in ['minute', 'hour', 'day', 'month', 'day_of_week']:
                j.pop(col, None)

        sdkapi.draw_dict(backend, "Scheduler List", jobs)
    except:
        traceback.print_exc()

@show.command("run-detail")
@click.argument("usecase")
@click.argument("pipeline")
@click.argument("runid")
@click.pass_context
def _run_detail(ctx, usecase, pipeline, runid):
    """
    Show detail of a given run
    """

    backend = ctx.obj["backend"]
    result = backend.run_detail(usecase, pipeline, runid)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    sdkapi.draw_run_detail(
        backend,
        "run-detail",
        result,
        {"Usecase": usecase, "Pipeline": pipeline, "RunID": runid},
    )


@show.command("tasks")
@click.argument("usecase")
@click.pass_context
def _tasks(ctx, usecase):
    """
    Show available tasks
    """

    backend = ctx.obj["backend"]
    result = backend.tasks(usecase)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    data = result["data"]
    for d in data:
        d["description"] = d["description"][:20]
        runs = sorted(d["runs"], reverse=True)
        if len(runs) > 0:
            d["lastrun"] = runs[0]
            d["runs"] = len(runs)
        else:
            d["lastrun"] = ""
            d["runs"] = 0

    columns = ["usecase", "name", "description", "path", "runs", "lastrun"]
    sdkapi.draw(
        backend,
        "tasks",
        result,
        columns,
        {
            "Usecase": usecase,
        },
    )


@show.command("task-run-list")
@click.argument("usecase")
@click.argument("task")
@click.pass_context
def _task_runs(ctx, usecase, task):
    """
    Show available runs for task
    """

    backend = ctx.obj["backend"]
    result = backend.task_run_list(usecase, task)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    result["data"] = sorted(result["data"], key=lambda r: r["start_time"], reverse=True)

    columns = ["runid", "status", "start_time", "end_time"]
    sdkapi.draw(
        backend, "task-run-list", result, columns, {"Usecase": usecase, "Task": task}
    )


@show.command("task-run-detail")
@click.argument("usecase")
@click.argument("task")
@click.argument("runid")
@click.pass_context
def _task_run_detail(ctx, usecase, task, runid):
    """
    Show detail of a given run
    """

    backend = ctx.obj["backend"]
    result = backend.task_run_detail(usecase, task, runid)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    sdkapi.draw_run_detail(
        backend,
        "run-detail",
        result,
        {"Usecase": usecase, "Task": task, "RunID": runid},
    )


@click.group()
@click.argument("backend")
@click.argument("service")
@click.pass_context
def featurestore(ctx, backend, service):
    """
    Access featurestore at backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    backends = config["backends"]

    relevant = None
    for b in backends:
        if (b.get("server", None) == backend) or (b.get("name", None) == backend):
            relevant = b
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))

    # Decode the key before accessing
    key = base64.b64decode(relevant["key"].encode("utf-8")).decode("utf-8")
    relevant["key"] = key

    backend = sdkapi.Backend(relevant)
    ctx.obj["backend"] = backend

    # Lookup the service as well..
    services = config["services"]
    relevant = None
    for s in services:
        if s.get("name", None) == service:
            relevant = s
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))
    ctx.obj["service"] = s


@featurestore.command("post")
@click.argument("filename")
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def _post(
    ctx,
    filename,
    debug,
):
    """
    Post json to server
    """

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    response = fslib.post(
        backend=backend, service=service, filename=filename, debug=debug
    )
    print(response)


@featurestore.command("generate")
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def _generate(ctx, debug):
    """
    Generate sample files
    """

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    response = fslib.generate(backend=backend, service=service, debug=debug)
    print(json.dumps(response, indent=4))


@featurestore.command("download")
@click.option("--featuregroup_id", default=None)
@click.option("--run_id", default=None)
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def _spec_download(
    ctx,
    featuregroup_id,
    run_id,
    debug,
):
    """
    Download a featuregroup specification
    """

    if featuregroup_id is None and run_id is None:
        raise Exception("One of the featuregroup_id or run_id should be specified")

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    response = fslib.download(
        backend=backend,
        service=service,
        featuregroup_id=featuregroup_id,
        run_id=run_id,
        debug=debug,
    )
    print(json.dumps(response, indent=4))


@featurestore.command(
    "search",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.option("--debug/--no-debug", default=False)
@click.argument("params", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def _spec_search(ctx, debug, params):
    """
    Search for a feature spec or run
    """

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    def error():
        print("Args should have attr=value format")
        print("Example: ....local spec search name=customer_persona")
        print("You can give django filter query name__icontains=persona")
        return

    for p in params:
        if p.count("=") != 1:
            return error()

    params = {p.split("=")[0]: p.split("=")[1] for p in params}
    if len(params) == 0:
        return error()

    response = fslib.search(
        backend=backend, service=service, params=params, debug=debug
    )

    print(json.dumps(response, indent=4))

#api.add_command(show)
#api.add_command(featurestore)
api.add_command(config)
api.add_command(_run_list)
api.add_command(_health)
api.add_command(scheduler)
api.add_command(_app_list)
api.add_command(_app_detail)

##########################################################
# Manage commands
##########################################################
@click.group("manage", hidden=True)
@click.option("--context", default=None, help="Environment file")
@click.pass_context
def manage(ctx, context):
    """
    Manage mongo/postgres etc.
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise

@manage.command("mongo")
@click.pass_context
def mongo_manage(ctx):
    """
    Start/stop mongo
    """

    commands = [
        "docker run --name alpha -d -p 27017:27017 mongo",
        "docker container stop alpha",
        "docker container ls"
        "docker container prune"
    ]

    for c in commands:
        print(c)


###################################
# Add to process
###################################
process.add_command(_version)
process.add_command(_start)
process.add_command(init_)
process.add_command(envsetup)
#process.add_command(api)
process.add_command(test)
process.add_command(doodle)
#process.add_command(sample)
process.add_command(manage)
process.add_command(_show_log)


def main():
    process()

if __name__ == "__main__":
    main()



import io
import os
import sys
import json
import time
import base64
import random
import logging
import hashlib
import traceback
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from enrichsdk import Compute, S3Mixin, EmailMixin
from enrichsdk.utils import note, SafeEncoder

from sklearn.neighbors import KNeighborsClassifier
from sklearn import svm
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import RocCurveDisplay, auc, roc_curve

from imblearn.over_sampling import SMOTE, RandomOverSampler
from imblearn.pipeline import make_pipeline

from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec
from enrichsdk.datasets import TransformDoodle

from dateutil import parser as dateparser

sns.set_context("notebook")

logger = logging.getLogger("app")

class ClassifierBase(Compute):
    """
    Take a training dataset and one or more eval datasets
    Builds a classification model using the training dataset
    Applies the model on the eval dataset(s) and generates predictions

    Features of transform baseclass include:
        * Flexible configuration
        * Highlevel specification of steps in ML classification flow:
            * specify multiple datasets (one for training, one or more for evaluation)
            * specify optional dataset prep methods
            * specify training model details with support for imbalanced datasets
            * specify evaluation strategy on one or more datasets
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "ClassifierBase"
        self.description = "Classification of data using a trained ML model"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

        self.epoch = time.time()    #for output path


    @classmethod
    def instantiable(cls):
        return False

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}

    ###########################################
    # Helper Functions
    ###########################################

    def update_frame(self, source, description, df, lineage=None):
        if isinstance(source, str):
            name = source
        else:
            name = source["name"]

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    def update_state(self, name, df, description, dependencies=[]):

        # append the existing dataframe if it exists
        if self.state.has_frame(name):
            frame = self.state.get_frame(name)
            df = pd.concat([df, frame['df']], ignore_index=True)

        updated_detail = {
            'df': df,
            'description': description,
            'transform': self.name,
            'frametype': 'pandas',
            'params': self.get_column_params(name, df) + dependencies
        }

        self.state.update_frame(name, updated_detail, create=True)


    ###########################################
    # I/O Functions
    ###########################################

    def read_s3_data(self, filename, params={}):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            try:
                df = pd.read_csv(s3.open(filename), **params)
                return df
            except:
                pass
        return None

    def get_dataset_s3(self, spec, source, paths, start_date, end_date):
        '''
        Gets all files from paths and puts them together
        into a single dataframe. If self.args['cache']==True,
        then this consolidated dataframe is cached / read from cache
        as applicable.
        '''
        msg = ""

        run_date    = self.args['run_date']
        config      = spec['config']
        dataset     = source['dataset']
        params      = source.get('params', {})

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{start_date}-to-{end_date}"
        cachefile = f"cache/{self.name}-rawdata-cache-" + cachename + ".csv"

        # read from cache if available
        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile, **params)
                logger.debug(f"Read cached {dataset}", extra={"transform": self.name, "data": msg})
                return df

        # read from S3
        dfs = []
        for path in paths:
            _df = self.read_s3_data(path, params)
            if _df is None:
                msg += f"Path error, skipping: {path}" + "\n"
                continue
            msg += f"Read from path: {path}" + "\n"
            dfs.append(_df)
        df = pd.concat(dfs)

        logger.debug(f"Read fresh {dataset}", extra={"transform": self.name})

        # Cache it for future use
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if (len(paths) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": paths,
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec, f"Dataset: {dataset}", df, lineage)

        return df

    def load_dataset(self, spec, name, source, datewindow):
        msg = ""

        source_id       = source['source_id']
        dataset         = source['dataset']
        source_version  = source.get('source_version', 'v1')

        start_date      = datewindow['start_date']
        end_date        = datewindow['end_date']

        if source_id == "custom":
            # we have a custom defined method in the derived class to generate the dataset
            if not hasattr(self, dataset):
                logger.exception(
                    f"No handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
                return None

            handler = getattr(self, dataset)
            params = source.get("params", {})
            df = handler(start_date, end_date, params)
            if df is None:
                logger.exception(
                    f"Couldn't load dataset for {spec['name']}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
                return None

            msg += f"Loaded using custom method: {dataset}" + "\n"
        else:
            # we are using the SDK to get the dataset
            datacred = self.args['datacred']
            doodle = TransformDoodle(self, self.state, datacred)

            d_source, paths = doodle.get_source_paths(start=start_date,
                                                    end=end_date,
                                                    name=dataset,
                                                    version=source_version,
                                                    source_id=source_id)

            msg += f"Paths: {paths}" + "\n"

            # Insert a read action..
            try:
                result = doodle.update_source(d_source['id'], {})
                logger.debug(f"Updated Doodle: {dataset}",
                             extra={
                                 'transform': self.name,
                                 'data': json.dumps(result, indent=4, cls=SafeEncoder)
                             })
            except:
                logger.exception(f"Unable to update Doodle {dataset}",
                                 extra={
                                     'transform': self.name
                                 })

            # read the data from S3
            df = self.get_dataset_s3(spec, source, paths, start_date, end_date)


        # check if we need to perform a post-processing step on the input data
        postprocess = source.get("postprocess")
        if postprocess != None:
            if not hasattr(self, postprocess['method']):
                logger.exception(
                    f"No post-process handler for: {dataset}",
                    extra={"transform": self.name, "data": json.dumps(source, indent=4)}
                )
            else:
                handler = getattr(self, postprocess['method'])
                params = postprocess.get("params", {})
                df = handler(df, params)
                msg += f"Post-processed dataset={dataset} using handler: {postprocess}" + "\n"

        msg += note(df, f"Input Dataset: {dataset}")
        logger.debug(
            f"Loaded dataset={dataset} for source={name}",
            extra={"transform": self.name, "data": msg}
        )

        return df

    def store_viz(self, spec, filename, viz, s3store=False):
        appname     = spec.get('app', self.name)
        name        = spec['name']
        namespace   = spec.get('namespace', 'default')

        msg = ""

        # store locally
        outputpath = os.path.join(self.args['output'], name)
        try:
            os.makedirs(outputpath)
        except:
            pass
        testplotfile = os.path.join(outputpath, f"{filename}")
        viz.savefig(testplotfile)
        msg += f"Stored (local): {testplotfile}" + "\n"

        # base64 encode the image
        img_data = io.BytesIO()
        viz.savefig(img_data, format='png')
        img_data.seek(0)
        img_data_b64 = base64.b64encode(img_data.read()).decode()

        # write to s3
        if s3store == True:
            # where are we storing it?
            run_date    = self.args['run_date']
            s3          = self.args['s3']
            epoch       = self.epoch
            targetdir   = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")
            vizfile     = os.path.join(targetdir, f"{filename}")

            img_data = io.BytesIO()
            viz.savefig(img_data, format='png')
            img_data.seek(0)
            with s3.open(vizfile, 'wb') as fd:
                fd.write(img_data.getbuffer())

            msg += f"Stored (remote): {vizfile}" + "\n"

        return msg

    def construct_result_metadata(self,):
        metadata = self.get_default_metadata(self.state)
        metadata['spec'] = spec
        metadata['predictspec'] = predictspec
        metadata['result'] = result
        return metadata


    def s3_store_result(self, spec, predictspec, result, df):
        appname     = spec.get('app',self.name)
        name        = spec['name']
        namespace   = spec.get('namespace', 'default')
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = self.epoch

        # where are we storing it?
        targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")

        predictspecname = predictspec.get("name", "NO-NAME")
        csvfile = os.path.join(targetdir, f"{predictspecname}-data.csv")
        resultfile = os.path.join(targetdir, f"{predictspecname}-result.json")

        # write to s3
        with s3.open(csvfile, 'w') as fd:
            df.to_csv(fd, index=False)
        with s3.open(resultfile, 'w') as fd:
            json.dump(result, fd, indent=4, cls=SafeEncoder)

        msg = f"Storage: S3" + "\n"
        msg += f"data CSV location: {csvfile}" + "\n"
        msg += f"result location: {resultfile}" + "\n"

        logger.debug(f"Wrote classification result [{predictspecname}]",
                        extra={"transform": self.name,
                                "data": msg})

    def disk_store_result(self, spec, predictspec, result, df):
        name = spec['name']
        predictspecname = predictspec['name']
        outputpath = os.path.join(self.args['output'], f"{name}")
        try:
            os.makedirs(outputpath)
        except:
            pass
        csvfile = os.path.join(outputpath, f"{predictspecname}-data.csv")
        resultfile = os.path.join(outputpath, f"{predictspecname}-result.json")

        msg = f"Storage: Disk" + "\n"
        msg += f"data CSV location: {csvfile}" + "\n"
        msg += f"result location: {resultfile}" + "\n"

        df.to_csv(csvfile, index=False)
        with open(resultfile, 'w') as fd:
            json.dump(result, fd, indent=4, cls=SafeEncoder)

        logger.debug(f"Wrote classification result [{predictspecname}]",
                        extra={"transform": self.name,
                                "data": msg})


    def load_sources(self, profilespec):
        """
        Load all the data sources
        """
        data = {}

        for source in profilespec.get('sources', []):
            name = source.get('name', 'NOT_SPECIFIED')

            # check for all fields needed
            if any(p not in source for p in ['nature', 'name', 'filename', 'stage']):
                logger.error(f"Malformed source [{name}]",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': json.dumps(source, indent=4)
                             }))
                continue

            if source['nature'] == 'disk':
                filename = source['filename']
                filename = f"{self.args['root']}/{filename}"
                df = pd.read_csv(filename)
                if df is None:
                    logger.error(f"Source not found [{name}]",
                                 extra=self.config.get_extra({
                                     'transform': self.name,
                                     'data': json.dumps(source, indent=4)
                                 }))
                    continue

                data[name] = df

            else:
                continue

            self.update_state(name, df, f"Source: {name}")

        # we have loaded all available data sources
        return data

    def store_result_data(self, spec, predictspec, result, df):
        name = predictspec.get("name")
        store   = predictspec.get('store', ["disk"])

        if "s3" in store:
            # store in s3
            self.s3_store_result(spec, predictspec, result, df)
        if "db" in store:
            # store in db
            self.db_store_result(spec, predictspec, result, df)
        if "disk" in store:
            # store in disk
            self.disk_store_result(spec, predictspec, result, df)

        # => Document the state
        self.update_state(name, df, f"Predictions for {name}")

    def store_metadata(self, spec, results):
        """
        Store all the metadata for the full run
        """
        metadata = self.get_default_metadata(self.state)
        metadata['spec'] = spec
        metadata['results'] = results

        store = spec.get("store", ["disk"])

        if "s3" in store:
            # store in s3
            appname     = spec.get('app',self.name)
            name        = spec['name']
            namespace   = spec.get('namespace', 'default')
            run_date    = self.args['run_date']
            s3          = self.args['s3']
            epoch       = self.epoch

            # where are we storing it?
            targetdir = os.path.join(self.args['s3root'], f"{appname}/{namespace}/{name}/{run_date}/{epoch}")
            metadatafile = os.path.join(targetdir, f"metadata.json")

            # write to s3
            with s3.open(metadatafile, 'w') as fd:
                json.dump(metadata, fd, indent=4, cls=SafeEncoder)
        if "db" in store:
            # store in db
            self.db_store_metadata(spec, predictspec, result, df)
        if "disk" in store:
            # store in disk
            name = spec['name']
            outfile = os.path.join(self.args['output'], f"{name}/metadata.json")
            with open(outfile, 'w') as fd:
                fd.write(json.dumps(metadata,indent=4))


    ##################################################################
    # Data preparation
    ##################################################################
    def prep_data(self, profilespec, data, artifacts):
        """
        Do any data prep needed
        We may need to do data scaling, normalization, etc. here
        Any artifacts of the prep that will be needed by the
        prediction stage must be returned in this function
        """

        # setup the training data and artifacts
        train_data = None
        for source in profilespec['sources']:
            if source['stage'] == "train":
                train_data = source['name']
        if train_data == None:
            return data, artifacts

        # check if a prep data method is specified
        prep_data = profilespec.get("prep", {}).get("method")
        if prep_data == None:
            return data, artifacts

        # call the prep data method
        msg = ""
        if hasattr(self, prep_data):
            handler = getattr(self, prep_data)
            data[train_data], artifacts, msg = handler(data[train_data], artifacts, 'train')

        logger.debug(f"Prepped training data [{train_data}]",
                     extra=self.config.get_extra({
                         'transform': self.name,
                         'data': msg
                     }))

        return data, artifacts

    ##################################################################
    # Training and prediction functions
    ##################################################################
    def make_predictions(self, profilespec, data, classifiers, artifacts):
        """
        Generate predictions for the various eval datasets
        """

        # to collect all the results
        results = {}

        # for each prediction spec in the profilespec
        for spec in profilespec.get('predict', []):
            # process it
            name = spec['name']
            if spec.get('enable', True) == False:
                logger.error(f"Spec [{name}] disabled, skipping",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                continue

            _dfs = []
            for source in spec.get('sources', []):
                _dfs.append(data[source])
            if len(_dfs)>0:
                eval_df = pd.concat(_dfs)
            else:
                logger.error(f"No sources to eval, skipping",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': json.dumps(spec, indent=4)
                             }))
                continue

            # get the target column name
            target = spec.get("target")
            ignore = spec.get("ignore", [])
            if target == None:
                logger.error(f"Target column not specified, skipping eval [{name}]",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': json.dumps(spec, indent=4)
                             }))
                continue

            # we now have eval_df
            # check if a prep data method is specified
            prep_data = profilespec.get("prep", {}).get("method")
            if prep_data != None:
                if hasattr(self, prep_data):
                    handler = getattr(self, prep_data)
                    prepped_eval_df, artifacts, msg = handler(eval_df, artifacts, 'predict')

                logger.debug(f"Prepped eval data [{name}]",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': msg
                             }))

            # check if all required columns are present
            missing_cols = set(artifacts['columns']).difference(set(prepped_eval_df.columns))
            for c in missing_cols:
                prepped_eval_df[c] = 0
            logger.debug(f"Added missing columns [{name}]",
                         extra=self.config.get_extra({
                             'transform': self.name,
                             'data': f"Missing cols: {missing_cols}"
                         }))


            # we now have the prepped eval df
            # run the specified classifier on it
            classifier_name = spec.get("model", "best")
            if classifier_name == "best":
                classifier_name = classifiers["best"]

            classifier = classifiers[classifier_name]['model']

            # create the data arrays
            X = prepped_eval_df[[c for c in prepped_eval_df.columns if c not in [target]+ignore]].to_numpy()

            # make the predictions
            r = classifier.predict(X)
            eval_df["__prediction"] = pd.Series(r)

            result = {
                "spec": spec,
                "n_datapoints": len(eval_df),
                "n_predictions": eval_df["__prediction"].value_counts().to_dict()
            }
            results[name] = result

            logger.debug(f"Predictions done [{name}]",
                         extra=self.config.get_extra({
                             'transform': self.name,
                             'data': note(eval_df, f"Predictions [{name}]")
                         }))

            # store results data csv
            self.store_result_data(profilespec, spec, result, eval_df)

        return results


    def get_classifier_pipeline(self, model):
        """
        Construct the classifier pipeline
            1. resampling
            2. classifier model
        """

        # do we need to resample the data
        # supports upsampling the minority class for now
        resample = model.get("resample")
        if resample == 'random':
            resampler = RandomOverSampler()
        elif resample == 'smote':
            resampler = SMOTE()
        else:
            # no resampling by default
            resampler = None

        # then get the classifier algorithm
        algorithm = model.get("model", {}).get("algorithm")
        params = model.get("model", {}).get("params", {})
        if algorithm == "knn":
            classifier = KNeighborsClassifier(**params)
        elif algorithm == "svm":
            classifier = svm.SVC(**params)
        else:
            # use the kNN algorithm by default
            classifier = KNeighborsClassifier(n_neighbors=3)

        # construct the pipeline
        if resampler == None:
            pipeline = classifier
        else:
            pipeline = make_pipeline(resampler, classifier)

        return pipeline

    def do_training(self, profilespec, modelspec, X, y, model, cv, metric):
        """
        Train a model given a dataset and a pipeline
        """

        msg = ""

        name = modelspec['name']

        tprs = []
        aucs = []
        mean_fpr = np.linspace(0, 1, 100)

        fig, ax = plt.subplots(figsize=(9, 9))
        for fold, (train, test) in enumerate(cv.split(X, y)):
            model.fit(X[train], y[train])
            viz = RocCurveDisplay.from_estimator(
                model,
                X[test],
                y[test],
                name=f"ROC fold {fold}",
                alpha=0.3,
                lw=1,
                ax=ax,
            )
            interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
            interp_tpr[0] = 0.0
            tprs.append(interp_tpr)
            aucs.append(viz.roc_auc)
        ax.plot([0, 1], [0, 1], "k--", label="chance level (AUC = 0.5)")

        # get the final model fit
        model.fit(X, y)

        mean_tpr = np.mean(tprs, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(aucs)
        ax.plot(
            mean_fpr,
            mean_tpr,
            color="b",
            label=r"Mean ROC (AUC = %0.2f $\pm$ %0.2f)" % (mean_auc, std_auc),
            lw=2,
            alpha=0.8,
        )

        std_tpr = np.std(tprs, axis=0)
        tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
        tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
        ax.fill_between(
            mean_fpr,
            tprs_lower,
            tprs_upper,
            color="grey",
            alpha=0.2,
            label=r"$\pm$ 1 std. dev.",
        )

        ax.set(
            xlim=[-0.05, 1.05],
            ylim=[-0.05, 1.05],
            xlabel="False Positive Rate",
            ylabel="True Positive Rate",
            title=f"[{name}] Mean ROC curve with variability')",
        )
        ax.axis("square")
        ax.legend(loc="lower right", fontsize=16)
        plt.tight_layout()

        # save training visualization
        filename = f"train-{name}-roc.png"
        l_msg = self.store_viz(profilespec, filename, plt)

        msg += l_msg

        # return the appropriate metric
        if metric == "auc":
            metric_val = mean_auc
        elif metric == "tpr":
            metric_val = mean_tpr
        elif metric == "fpr":
            metric_val = mean_fpr
        else:
            metric_val = mean_auc

        classifier = {
            "model": model,
            "metric": metric_val
        }

        return classifier, msg

    def decide_best_classifier(self, classifiers):
        max_metric = 0
        for name, classifier in classifiers.items():
            if classifier['metric'] >= max_metric:
                best = name
                max_metric = classifier['metric']
        return best



    def train_models(self, profilespec, data):
        """
        Model training
        """
        msg = ""

        # prep the training data
        # and generate any artifacts needed later
        artifacts = {}
        data, artifacts = self.prep_data(profilespec, data, artifacts)
        # we need a list of all columns which will be used in training
        for source in profilespec['sources']:
            if source['stage'] == "train":
                train_data = source['name']
                artifacts['columns'] = list(data[train_data].columns)

        # required params
        trainspec = profilespec.get("train")
        metric  = trainspec.get("metric", "auc") #what is the metric against which to compare models
        folds   = trainspec.get("folds", 1)      #how many folds for cross validation

        classifiers = {}

        # for each model to train
        models = trainspec.get("models", [])
        for model in models:
            if model.get("enable", True) == False:
                continue

            name    = model.get("name", f"{hashlib.md5(json.dumps(model).encode('utf-8')).hexdigest()}")
            model['name'] = name
            dataset = model.get("source")
            target  = model.get("target")
            ignore  = model.get("ignore", [])

            if dataset == None or dataset not in data:
                logger.error(f"Dataset not known, skipping training [{name}]",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': json.dumps(model, indent=4)
                             }))
                continue
            if target == None:
                logger.error(f"Target column not specified, skipping training [{name}]",
                             extra=self.config.get_extra({
                                 'transform': self.name,
                                 'data': json.dumps(model, indent=4)
                             }))
                continue

            msg += f"Model: {name}" + "\n"
            msg += f"Dataset: {dataset}" + "\n"
            msg += f"Target column: {target}" + "\n"
            msg += f"Ignore columns: {ignore}" + "\n"

            df = data[dataset]

            # create the data arrays
            X = df[[c for c in df.columns if c not in [target]+ignore]].to_numpy()
            y = df[target].to_numpy()

            msg += f"Size (X): {X.size}" + "\n"
            msg += f"Size (y): {y.size}" + "\n"

            # figure out the minority class
            # in case we need to resample
            class_distribution = pd.Series(y).value_counts(normalize=True)
            pos_label = class_distribution.idxmin()
            msg += f"Positive label: {pos_label}" + "\n"

            # construct the classifier pipeline object
            classifier_pipeline = self.get_classifier_pipeline(model)

            # set up the n-fold cross validation
            cv = StratifiedKFold(n_splits=folds)

            # do model training
            classifiers[name], l_msg = self.do_training(profilespec, model, X, y, classifier_pipeline, cv, metric)
            msg += l_msg

        # decide on what the best classifier is based on the metric
        classifiers['best'] = self.decide_best_classifier(classifiers)

        msg += f"Classifiers: {json.dumps(classifiers, indent=4, cls=SafeEncoder)}" + "\n"
        msg += f"Artifacts: {json.dumps(artifacts, indent=4, cls=SafeEncoder)}" + "\n"

        logger.debug(f"Completed training",
                     extra=self.config.get_extra({
                         'transform': self.name,
                         'data': msg
                     }))

        return classifiers, artifacts


    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug("Start execution",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))
        self.state = state

        # Get the profile spec
        is_valid, profile, msg = profilespec.get_profile(self, "policyapp.classifier")
        if is_valid:
            logger.debug(
                f"Loaded profilespec",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")

        specs = profile.get("specs", None)
        if specs is None:
            raise Exception("Could not find 'specs' in profile")

        # Now go through each spec and process it
        for spec in specs:

            ###
            # first, some checks on the spec
            do_process_spec = True
            name = spec.get('name', 'NO_SPEC_NAME')

            enabled = spec.get("active", True)
            if not enabled:
                logger.debug(
                    f"Spec not enabled, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                do_process_spec = False
                continue

            for f in ["name", "sources", "train", "predict"]:
                if f not in spec:
                    logger.exception(
                        f"Spec has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            ###
            # get all the data sources
            ##############
            # Re-write to use standard load_dataset(...) method
            # when using API-based spec and s3 data sources
            data = self.load_sources(spec)
            ##############
            if len(data) == 0:
                logger.exception("No datasources found, failing",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("No datasources")

            ###
            # model training stage
            classifiers, artifacts = self.train_models(spec, data)

            ###
            # make predictions for each evaluation dataset
            # and store results
            results = self.make_predictions(spec, data, classifiers, artifacts)

            # Store the metadata with results
            self.store_metadata(spec, results)

        # Done
        logger.debug("Complete execution",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass
